import os
import re
import tempfile
from pathlib import Path

from commitizen import git
from commitizen.cz import exceptions

_RE_LOCAL_VERSION = re.compile(r"\+.+")


def required_validator(answer: str, msg: object = None) -> str:
    if not answer:
        raise exceptions.AnswerRequiredError(msg)
    return answer


def multiple_line_breaker(answer: str, sep: str = "|") -> str:
    return "\n".join(line.strip() for line in answer.split(sep) if line)


def strip_local_version(version: str) -> str:
    return _RE_LOCAL_VERSION.sub("", version)


def get_backup_file_path() -> Path:
    project_root = git.find_git_project_root()
    project = project_root.as_posix().replace("/", "%") if project_root else ""

    user = os.environ.get("USER", "")
    return Path(tempfile.gettempdir(), f"cz.commit%{user}%{project}.backup")



import os
from unittest.mock import ANY

import pytest
from pytest_mock import MockFixture

from commitizen import cmd, commands
from commitizen.cz.exceptions import CzException
from commitizen.cz.utils import get_backup_file_path
from commitizen.exceptions import (
    CommitError,
    CommitMessageLengthExceededError,
    CustomError,
    DryRunExit,
    NoAnswersError,
    NoCommitBackupError,
    NotAGitProjectError,
    NotAllowed,
    NothingToCommitError,
)


@pytest.fixture
def staging_is_clean(mocker: MockFixture, tmp_git_project):
    is_staging_clean_mock = mocker.patch("commitizen.git.is_staging_clean")
    is_staging_clean_mock.return_value = False
    return tmp_git_project


@pytest.fixture
def backup_file(tmp_git_project):
    with open(get_backup_file_path(), "w") as backup_file:
        backup_file.write("backup commit")


@pytest.mark.usefixtures("staging_is_clean")
def test_commit(config, mocker: MockFixture):
    prompt_mock = mocker.patch("questionary.prompt")
    prompt_mock.return_value = {
        "prefix": "feat",
        "subject": "user created",
        "scope": "",
        "is_breaking_change": False,
        "body": "",
        "footer": "",
    }

    commit_mock = mocker.patch("commitizen.git.commit")
    commit_mock.return_value = cmd.Command("success", "", b"", b"", 0)
    success_mock = mocker.patch("commitizen.out.success")

    commands.Commit(config, {})()
    success_mock.assert_called_once()


@pytest.mark.usefixtures("staging_is_clean")
def test_commit_backup_on_failure(config, mocker: MockFixture):
    prompt_mock = mocker.patch("questionary.prompt")
    prompt_mock.return_value = {
        "prefix": "feat",
        "subject": "user created",
        "scope": "",
        "is_breaking_change": False,
        "body": "closes #21",
        "footer": "",
    }

    commit_mock = mocker.patch("commitizen.git.commit")
    commit_mock.return_value = cmd.Command("", "error", b"", b"", 9)
    error_mock = mocker.patch("commitizen.out.error")

    with pytest.raises(CommitError):
        commit_cmd = commands.Commit(config, {})
        temp_file = commit_cmd.backup_file_path
        commit_cmd()

    prompt_mock.assert_called_once()
    error_mock.assert_called_once()
    assert os.path.isfile(temp_file)


@pytest.mark.usefixtures("staging_is_clean")
def test_commit_retry_fails_no_backup(config, mocker: MockFixture):
    commit_mock = mocker.patch("commitizen.git.commit")
    commit_mock.return_value = cmd.Command("success", "", b"", b"", 0)

    with pytest.raises(NoCommitBackupError) as excinfo:
        commands.Commit(config, {"retry": True})()

    assert NoCommitBackupError.message in str(excinfo.value)


@pytest.mark.usefixtures("staging_is_clean", "backup_file")
def test_commit_retry_works(config, mocker: MockFixture):
    prompt_mock = mocker.patch("questionary.prompt")

    commit_mock = mocker.patch("commitizen.git.commit")
    commit_mock.return_value = cmd.Command("success", "", b"", b"", 0)
    success_mock = mocker.patch("commitizen.out.success")

    commit_cmd = commands.Commit(config, {"retry": True})
    temp_file = commit_cmd.backup_file_path
    commit_cmd()

    commit_mock.assert_called_with("backup commit", args="")
    prompt_mock.assert_not_called()
    success_mock.assert_called_once()
    assert not os.path.isfile(temp_file)


@pytest.mark.usefixtures("staging_is_clean")
def test_commit_retry_after_failure_no_backup(config, mocker: MockFixture):
    prompt_mock = mocker.patch("questionary.prompt")
    prompt_mock.return_value = {
        "prefix": "feat",
        "subject": "user created",
        "scope": "",
        "is_breaking_change": False,
        "body": "closes #21",
        "footer": "",
    }

    commit_mock = mocker.patch("commitizen.git.commit")
    commit_mock.return_value = cmd.Command("success", "", b"", b"", 0)
    success_mock = mocker.patch("commitizen.out.success")

    config.settings["retry_after_failure"] = True
    commands.Commit(config, {})()

    commit_mock.assert_called_with("feat: user created\n\ncloses #21", args="")
    prompt_mock.assert_called_once()
    success_mock.assert_called_once()


@pytest.mark.usefixtures("staging_is_clean", "backup_file")
def test_commit_retry_after_failure_works(config, mocker: MockFixture):
    prompt_mock = mocker.patch("questionary.prompt")

    commit_mock = mocker.patch("commitizen.git.commit")
    commit_mock.return_value = cmd.Command("success", "", b"", b"", 0)
    success_mock = mocker.patch("commitizen.out.success")

    config.settings["retry_after_failure"] = True
    commit_cmd = commands.Commit(config, {})
    temp_file = commit_cmd.backup_file_path
    commit_cmd()

    commit_mock.assert_called_with("backup commit", args="")
    prompt_mock.assert_not_called()
    success_mock.assert_called_once()
    assert not os.path.isfile(temp_file)


@pytest.mark.usefixtures("staging_is_clean", "backup_file")
def test_commit_retry_after_failure_with_no_retry_works(config, mocker: MockFixture):
    prompt_mock = mocker.patch("questionary.prompt")
    prompt_mock.return_value = {
        "prefix": "feat",
        "subject": "user created",
        "scope": "",
        "is_breaking_change": False,
        "body": "closes #21",
        "footer": "",
    }

    commit_mock = mocker.patch("commitizen.git.commit")
    commit_mock.return_value = cmd.Command("success", "", b"", b"", 0)
    success_mock = mocker.patch("commitizen.out.success")

    config.settings["retry_after_failure"] = True
    commit_cmd = commands.Commit(config, {"no_retry": True})
    temp_file = commit_cmd.backup_file_path
    commit_cmd()

    commit_mock.assert_called_with("feat: user created\n\ncloses #21", args="")
    prompt_mock.assert_called_once()
    success_mock.assert_called_once()
    assert not os.path.isfile(temp_file)


@pytest.mark.usefixtures("staging_is_clean")
def test_commit_command_with_dry_run_option(config, mocker: MockFixture):
    prompt_mock = mocker = mocker.patch("questionary.prompt")
    prompt_mock.return_value = {
        "prefix": "feat",
        "subject": "user created",
        "scope": "",
        "is_breaking_change": False,
        "body": "closes #57",
        "footer": "",
    }

    with pytest.raises(DryRunExit):
        commit_cmd = commands.Commit(config, {"dry_run": True})
        commit_cmd()


@pytest.mark.usefixtures("staging_is_clean")
def test_commit_command_with_write_message_to_file_option(
    config, tmp_path, mocker: MockFixture
):
    tmp_file = tmp_path / "message"

    prompt_mock = mocker.patch("questionary.prompt")
    prompt_mock.return_value = {
        "prefix": "feat",
        "subject": "user created",
        "scope": "",
        "is_breaking_change": False,
        "body": "",
        "footer": "",
    }

    commit_mock = mocker.patch("commitizen.git.commit")
    commit_mock.return_value = cmd.Command("success", "", b"", b"", 0)
    success_mock = mocker.patch("commitizen.out.success")

    commands.Commit(config, {"write_message_to_file": tmp_file})()
    success_mock.assert_called_once()
    assert tmp_file.exists()
    assert tmp_file.read_text() == "feat: user created"


@pytest.mark.usefixtures("staging_is_clean")
def test_commit_command_with_invalid_write_message_to_file_option(
    config, tmp_path, mocker: MockFixture
):
    prompt_mock = mocker.patch("questionary.prompt")
    prompt_mock.return_value = {
        "prefix": "feat",
        "subject": "user created",
        "scope": "",
        "is_breaking_change": False,
        "body": "",
        "footer": "",
    }

    with pytest.raises(NotAllowed):
        commit_cmd = commands.Commit(config, {"write_message_to_file": tmp_path})
        commit_cmd()


@pytest.mark.usefixtures("staging_is_clean")
def test_commit_command_with_signoff_option(config, mocker: MockFixture):
    prompt_mock = mocker.patch("questionary.prompt")
    prompt_mock.return_value = {
        "prefix": "feat",
        "subject": "user created",
        "scope": "",
        "is_breaking_change": False,
        "body": "",
        "footer": "",
    }

    commit_mock = mocker.patch("commitizen.git.commit")
    commit_mock.return_value = cmd.Command("success", "", b"", b"", 0)
    success_mock = mocker.patch("commitizen.out.success")

    commands.Commit(config, {"signoff": True})()

    commit_mock.assert_called_once_with(ANY, args="-s")
    success_mock.assert_called_once()


@pytest.mark.usefixtures("staging_is_clean")
def test_commit_command_with_always_signoff_enabled(config, mocker: MockFixture):
    prompt_mock = mocker.patch("questionary.prompt")
    prompt_mock.return_value = {
        "prefix": "feat",
        "subject": "user created",
        "scope": "",
        "is_breaking_change": False,
        "body": "",
        "footer": "",
    }

    commit_mock = mocker.patch("commitizen.git.commit")
    commit_mock.return_value = cmd.Command("success", "", b"", b"", 0)
    success_mock = mocker.patch("commitizen.out.success")

    config.settings["always_signoff"] = True
    commands.Commit(config, {})()

    commit_mock.assert_called_once_with(ANY, args="-s")
    success_mock.assert_called_once()


@pytest.mark.usefixtures("staging_is_clean")
def test_commit_command_with_gpgsign_and_always_signoff_enabled(
    config, mocker: MockFixture
):
    prompt_mock = mocker.patch("questionary.prompt")
    prompt_mock.return_value = {
        "prefix": "feat",
        "subject": "user created",
        "scope": "",
        "is_breaking_change": False,
        "body": "",
        "footer": "",
    }

    commit_mock = mocker.patch("commitizen.git.commit")
    commit_mock.return_value = cmd.Command("success", "", b"", b"", 0)
    success_mock = mocker.patch("commitizen.out.success")

    config.settings["always_signoff"] = True
    commands.Commit(config, {"extra_cli_args": "-S"})()

    commit_mock.assert_called_once_with(ANY, args="-S -s")
    success_mock.assert_called_once()


@pytest.mark.usefixtures("tmp_git_project")
def test_commit_when_nothing_to_commit(config, mocker: MockFixture):
    is_staging_clean_mock = mocker.patch("commitizen.git.is_staging_clean")
    is_staging_clean_mock.return_value = True

    with pytest.raises(NothingToCommitError) as excinfo:
        commit_cmd = commands.Commit(config, {})
        commit_cmd()

    assert "No files added to staging!" in str(excinfo.value)


@pytest.mark.usefixtures("staging_is_clean")
def test_commit_with_allow_empty(config, mocker: MockFixture):
    prompt_mock = mocker.patch("questionary.prompt")
    prompt_mock.return_value = {
        "prefix": "feat",
        "subject": "user created",
        "scope": "",
        "is_breaking_change": False,
        "body": "closes #21",
        "footer": "",
    }

    commit_mock = mocker.patch("commitizen.git.commit")
    commit_mock.return_value = cmd.Command("success", "", b"", b"", 0)
    success_mock = mocker.patch("commitizen.out.success")

    commands.Commit(config, {"extra_cli_args": "--allow-empty"})()

    commit_mock.assert_called_with(
        "feat: user created\n\ncloses #21", args="--allow-empty"
    )
    success_mock.assert_called_once()


@pytest.mark.usefixtures("staging_is_clean")
def test_commit_with_signoff_and_allow_empty(config, mocker: MockFixture):
    prompt_mock = mocker.patch("questionary.prompt")
    prompt_mock.return_value = {
        "prefix": "feat",
        "subject": "user created",
        "scope": "",
        "is_breaking_change": False,
        "body": "closes #21",
        "footer": "",
    }

    commit_mock = mocker.patch("commitizen.git.commit")
    commit_mock.return_value = cmd.Command("success", "", b"", b"", 0)
    success_mock = mocker.patch("commitizen.out.success")

    config.settings["always_signoff"] = True
    commands.Commit(config, {"extra_cli_args": "--allow-empty"})()

    commit_mock.assert_called_with(
        "feat: user created\n\ncloses #21", args="--allow-empty -s"
    )
    success_mock.assert_called_once()


@pytest.mark.usefixtures("staging_is_clean")
def test_commit_when_customized_expected_raised(config, mocker: MockFixture, capsys):
    _err = ValueError()
    _err.__context__ = CzException("This is the root custom err")
    prompt_mock = mocker.patch("questionary.prompt")
    prompt_mock.side_effect = _err

    with pytest.raises(CustomError) as excinfo:
        commit_cmd = commands.Commit(config, {})
        commit_cmd()

    # Assert only the content in the formatted text
    assert "This is the root custom err" in str(excinfo.value)


@pytest.mark.usefixtures("staging_is_clean")
def test_commit_when_non_customized_expected_raised(
    config, mocker: MockFixture, capsys
):
    _err = ValueError()
    prompt_mock = mocker.patch("questionary.prompt")
    prompt_mock.side_effect = _err

    with pytest.raises(ValueError):
        commit_cmd = commands.Commit(config, {})
        commit_cmd()


@pytest.mark.usefixtures("staging_is_clean")
def test_commit_when_no_user_answer(config, mocker: MockFixture, capsys):
    prompt_mock = mocker.patch("questionary.prompt")
    prompt_mock.return_value = None

    with pytest.raises(NoAnswersError):
        commit_cmd = commands.Commit(config, {})
        commit_cmd()


def test_commit_in_non_git_project(tmpdir, config):
    with tmpdir.as_cwd():
        with pytest.raises(NotAGitProjectError):
            commands.Commit(config, {})


@pytest.mark.usefixtures("staging_is_clean")
def test_commit_command_with_all_option(config, mocker: MockFixture):
    prompt_mock = mocker.patch("questionary.prompt")
    prompt_mock.return_value = {
        "prefix": "feat",
        "subject": "user created",
        "scope": "",
        "is_breaking_change": False,
        "body": "",
        "footer": "",
    }

    commit_mock = mocker.patch("commitizen.git.commit")
    commit_mock.return_value = cmd.Command("success", "", b"", b"", 0)
    success_mock = mocker.patch("commitizen.out.success")
    add_mock = mocker.patch("commitizen.git.add")
    commands.Commit(config, {"all": True})()
    add_mock.assert_called()
    success_mock.assert_called_once()


@pytest.mark.usefixtures("staging_is_clean")
def test_commit_command_with_extra_args(config, mocker: MockFixture):
    prompt_mock = mocker.patch("questionary.prompt")
    prompt_mock.return_value = {
        "prefix": "feat",
        "subject": "user created",
        "scope": "",
        "is_breaking_change": False,
        "body": "",
        "footer": "",
    }

    commit_mock = mocker.patch("commitizen.git.commit")
    commit_mock.return_value = cmd.Command("success", "", b"", b"", 0)
    success_mock = mocker.patch("commitizen.out.success")
    commands.Commit(config, {"extra_cli_args": "-- -extra-args1 -extra-arg2"})()
    commit_mock.assert_called_once_with(ANY, args="-- -extra-args1 -extra-arg2")
    success_mock.assert_called_once()


@pytest.mark.usefixtures("staging_is_clean")
def test_commit_command_with_message_length_limit(config, mocker: MockFixture):
    prompt_mock = mocker.patch("questionary.prompt")
    prefix = "feat"
    subject = "random subject"
    message_length = len(prefix) + len(": ") + len(subject)
    prompt_mock.return_value = {
        "prefix": prefix,
        "subject": subject,
        "scope": "",
        "is_breaking_change": False,
        "body": "random body",
        "footer": "random footer",
    }

    commit_mock = mocker.patch("commitizen.git.commit")
    commit_mock.return_value = cmd.Command("success", "", b"", b"", 0)
    success_mock = mocker.patch("commitizen.out.success")

    commands.Commit(config, {"message_length_limit": message_length})()
    success_mock.assert_called_once()

    with pytest.raises(CommitMessageLengthExceededError):
        commands.Commit(config, {"message_length_limit": message_length - 1})()


@pytest.mark.usefixtures("staging_is_clean")
@pytest.mark.parametrize("editor", ["vim", None])
def test_manual_edit(editor, config, mocker: MockFixture, tmp_path):
    mocker.patch("commitizen.git.get_core_editor", return_value=editor)
    subprocess_mock = mocker.patch("subprocess.call")

    mocker.patch("shutil.which", return_value=editor)

    test_message = "Initial commit message"
    temp_file = tmp_path / "temp_commit_message"
    temp_file.write_text(test_message)

    mock_temp_file = mocker.patch("tempfile.NamedTemporaryFile")
    mock_temp_file.return_value.__enter__.return_value.name = str(temp_file)

    commit_cmd = commands.Commit(config, {"edit": True})

    if editor is None:
        with pytest.raises(RuntimeError):
            commit_cmd.manual_edit(test_message)
    else:
        edited_message = commit_cmd.manual_edit(test_message)

        subprocess_mock.assert_called_once_with(["vim", str(temp_file)])

        assert edited_message == test_message.strip()


@pytest.mark.usefixtures("staging_is_clean")
@pytest.mark.parametrize(
    "out", ["no changes added to commit", "nothing added to commit"]
)
def test_commit_when_nothing_added_to_commit(config, mocker: MockFixture, out):
    prompt_mock = mocker.patch("questionary.prompt")
    prompt_mock.return_value = {
        "prefix": "feat",
        "subject": "user created",
        "scope": "",
        "is_breaking_change": False,
        "body": "",
        "footer": "",
    }

    commit_mock = mocker.patch("commitizen.git.commit")
    commit_mock.return_value = cmd.Command(
        out=out,
        err="",
        stdout=out.encode(),
        stderr=b"",
        return_code=0,
    )
    error_mock = mocker.patch("commitizen.out.error")

    commands.Commit(config, {})()

    commit_mock.assert_called_once()
    error_mock.assert_called_once_with(out)


@pytest.mark.usefixtures("staging_is_clean")
def test_commit_command_with_config_message_length_limit(config, mocker: MockFixture):
    prompt_mock = mocker.patch("questionary.prompt")
    prefix = "feat"
    subject = "random subject"
    message_length = len(prefix) + len(": ") + len(subject)
    prompt_mock.return_value = {
        "prefix": prefix,
        "subject": subject,
        "scope": "",
        "is_breaking_change": False,
        "body": "random body",
        "footer": "random footer",
    }

    commit_mock = mocker.patch("commitizen.git.commit")
    commit_mock.return_value = cmd.Command("success", "", b"", b"", 0)
    success_mock = mocker.patch("commitizen.out.success")

    config.settings["message_length_limit"] = message_length
    commands.Commit(config, {})()
    success_mock.assert_called_once()

    config.settings["message_length_limit"] = message_length - 1
    with pytest.raises(CommitMessageLengthExceededError):
        commands.Commit(config, {})()


@pytest.mark.usefixtures("staging_is_clean")
def test_commit_command_cli_overrides_config_message_length_limit(
    config, mocker: MockFixture
):
    prompt_mock = mocker.patch("questionary.prompt")
    prefix = "feat"
    subject = "random subject"
    message_length = len(prefix) + len(": ") + len(subject)
    prompt_mock.return_value = {
        "prefix": prefix,
        "subject": subject,
        "scope": "",
        "is_breaking_change": False,
        "body": "random body",
        "footer": "random footer",
    }

    commit_mock = mocker.patch("commitizen.git.commit")
    commit_mock.return_value = cmd.Command("success", "", b"", b"", 0)
    success_mock = mocker.patch("commitizen.out.success")

    config.settings["message_length_limit"] = message_length - 1

    commands.Commit(config, {"message_length_limit": message_length})()
    success_mock.assert_called_once()

    success_mock.reset_mock()
    commands.Commit(config, {"message_length_limit": None})()
    success_mock.assert_called_once()



import pytest
from pytest_mock import MockFixture

from commitizen.cz import exceptions, utils


def test_required_validator():
    assert utils.required_validator("test") == "test"

    with pytest.raises(exceptions.AnswerRequiredError):
        utils.required_validator("")


def test_multiple_line_breaker():
    message = "this is the first line    | and this is the second line   "
    result = utils.multiple_line_breaker(message)
    assert result == "this is the first line\nand this is the second line"

    result = utils.multiple_line_breaker(message, "is")
    assert result == "th\n\nthe first line    | and th\n\nthe second line"


def test_get_backup_file_path_no_project_root(mocker: MockFixture):
    project_root_mock = mocker.patch("commitizen.git.find_git_project_root")
    project_root_mock.return_value = None
    assert utils.get_backup_file_path()



import os

import pytest

from commitizen import defaults
from commitizen.config import BaseConfig
from commitizen.config.json_config import JsonConfig


@pytest.fixture()
def config():
    _config = BaseConfig()
    _config.settings.update({"name": defaults.DEFAULT_SETTINGS["name"]})
    return _config


@pytest.fixture()
def config_customize():
    json_string = r"""{
      "commitizen": {
        "name": "cz_customize",
        "version": "3.0.0",
        "changelog_incremental": "true",
        "customize": {
          "message_template": "{{prefix}}({{scope}}): {{subject}}\n\n{{body}}{% if is_breaking_change %}\nBREAKING CHANGE: {{footer}}{% endif %}",
          "schema": "<type>(<scope>): <subject>\n<BLANK LINE>\n<body>\n<BLANK LINE>\n(BREAKING CHANGE: <footer>)",
          "schema_pattern": "(build|ci|docs|feat|fix|perf|refactor|style|test|chore|revert|bump)(\\(\\S+\\))?!?:(\\s.*)",
          "change_type_map": {
            "feat": "Feat",
            "fix": "Fix",
            "refactor": "Refactor",
            "perf": "Perf"
          },
          "change_type_order": ["Refactor", "Feat"],
          "commit_parser": "^(?P<change_type>feat|fix|refactor|perf|BREAKING CHANGE)(?:\\((?P<scope>[^()\\r\\n]*)\\)|\\()?(?P<breaking>!)?:\\s(?P<message>.*)?",
          "changelog_pattern": "^(BREAKING[\\-\\ ]CHANGE|feat|fix|refactor|perf)(\\(.+\\))?(!)?",
          "questions": [

          ]
        }
      }
    }"""
    _config = JsonConfig(data=json_string, path="not_exist.json")
    return _config


@pytest.fixture()
def changelog_path() -> str:
    return os.path.join(os.getcwd(), "CHANGELOG.md")


@pytest.fixture()
def config_path() -> str:
    return os.path.join(os.getcwd(), "pyproject.toml")



from __future__ import annotations

import re
import warnings
from dataclasses import dataclass, field
from functools import cached_property
from itertools import chain
from string import Template
from typing import TYPE_CHECKING, NamedTuple

from commitizen import out
from commitizen.defaults import DEFAULT_SETTINGS, Settings, get_tag_regexes
from commitizen.git import GitTag
from commitizen.version_schemes import (
    DEFAULT_SCHEME,
    InvalidVersion,
    Version,
    VersionScheme,
    get_version_scheme,
)

if TYPE_CHECKING:
    import sys
    from collections.abc import Iterable, Sequence

    from commitizen.version_schemes import VersionScheme

    # Self is Python 3.11+ but backported in typing-extensions
    if sys.version_info < (3, 11):
        from typing_extensions import Self
    else:
        from typing import Self


class VersionTag(NamedTuple):
    """Represent a version and its matching tag form."""

    version: str
    tag: str


@dataclass
class TagRules:
    """
    Encapsulate tag-related rules.

    It allows to filter or match tags according to rules provided in settings:
    - `tag_format`: the current format of the tags generated on `bump`
    - `legacy_tag_formats`: previous known formats of the tag
    - `ignored_tag_formats`: known formats that should be ignored
    - `merge_prereleases`: if `True`, prereleases will be merged with their release counterpart
    - `version_scheme`: the version scheme to use, which will be used to parse and format versions

    This class is meant to abstract and centralize all the logic related to tags.
    To ensure consistency, it is recommended to use this class to handle tags.

    Example:

    ```python
    settings = DEFAULT_SETTINGS.clone()
    settings.update(
        {
            "tag_format": "v{version}",
            "legacy_tag_formats": ["version{version}", "ver{version}"],
            "ignored_tag_formats": ["ignored{version}"],
        }
    )

    rules = TagRules.from_settings(settings)

    assert rules.is_version_tag("v1.0.0")
    assert rules.is_version_tag("version1.0.0")
    assert rules.is_version_tag("ver1.0.0")
    assert not rules.is_version_tag("ignored1.0.0", warn=True)  # Does not warn
    assert not rules.is_version_tag("warn1.0.0", warn=True)  # Does warn

    assert rules.search_version("# My v1.0.0 version").version == "1.0.0"
    assert rules.extract_version("v1.0.0") == Version("1.0.0")
    try:
        assert rules.extract_version("not-a-v1.0.0")
    except InvalidVersion:
        print("Does not match a tag format")
    ```
    """

    scheme: VersionScheme = DEFAULT_SCHEME
    tag_format: str = DEFAULT_SETTINGS["tag_format"]
    legacy_tag_formats: Sequence[str] = field(default_factory=list)
    ignored_tag_formats: Sequence[str] = field(default_factory=list)
    merge_prereleases: bool = False

    @property
    def tag_formats(self) -> Iterable[str]:
        return chain([self.tag_format], self.legacy_tag_formats)

    @cached_property
    def version_regexes(self) -> list[re.Pattern]:
        """Regexes for all legit tag formats, current and legacy"""
        return [re.compile(self._format_regex(f)) for f in self.tag_formats]

    @cached_property
    def ignored_regexes(self) -> list[re.Pattern]:
        """Regexes for known but ignored tag formats"""
        return [
            re.compile(self._format_regex(f, star=True))
            for f in self.ignored_tag_formats
        ]

    def _format_regex(self, tag_pattern: str, star: bool = False) -> str:
        """
        Format a tag pattern into a regex pattern.

        If star is `True`, the `*` character will be considered as a wildcard.
        """
        tag_regexes = get_tag_regexes(self.scheme.parser.pattern)
        format_regex = tag_pattern.replace("*", "(?:.*?)") if star else tag_pattern
        for pattern, regex in tag_regexes.items():
            format_regex = format_regex.replace(pattern, regex)
        return format_regex

    def _version_tag_error(self, tag: str) -> str:
        """Format the error message for an invalid version tag"""
        return f"Invalid version tag: '{tag}' does not match any configured tag format"

    def is_version_tag(self, tag: str | GitTag, warn: bool = False) -> bool:
        """
        True if a given tag is a legit version tag.

        if `warn` is `True`, it will print a warning message if the tag is not a version tag.
        """
        tag = tag.name if isinstance(tag, GitTag) else tag
        is_legit = any(regex.fullmatch(tag) for regex in self.version_regexes)
        if warn and not is_legit and not self.is_ignored_tag(tag):
            out.warn(self._version_tag_error(tag))
        return is_legit

    def is_ignored_tag(self, tag: str | GitTag) -> bool:
        """True if a given tag can be ignored"""
        tag = tag.name if isinstance(tag, GitTag) else tag
        return any(regex.match(tag) for regex in self.ignored_regexes)

    def get_version_tags(
        self, tags: Iterable[GitTag], warn: bool = False
    ) -> list[GitTag]:
        """Filter in version tags and warn on unexpected tags"""
        return [tag for tag in tags if self.is_version_tag(tag, warn)]

    def extract_version(self, tag: GitTag) -> Version:
        """
        Extract a version from the tag as defined in tag formats.

        Raises `InvalidVersion` if the tag does not match any format.
        """
        candidates = (
            m for regex in self.version_regexes if (m := regex.fullmatch(tag.name))
        )
        if not (m := next(candidates, None)):
            raise InvalidVersion(self._version_tag_error(tag.name))
        if "version" in m.groupdict():
            return self.scheme(m.group("version"))

        parts = m.groupdict()
        version = parts["major"]

        if minor := parts.get("minor"):
            version = f"{version}.{minor}"
        if patch := parts.get("patch"):
            version = f"{version}.{patch}"

        if parts.get("prerelease"):
            version = f"{version}-{parts['prerelease']}"
        if parts.get("devrelease"):
            version = f"{version}{parts['devrelease']}"
        return self.scheme(version)

    def include_in_changelog(self, tag: GitTag) -> bool:
        """Check if a tag should be included in the changelog"""
        try:
            version = self.extract_version(tag)
        except InvalidVersion:
            return False
        return not (self.merge_prereleases and version.is_prerelease)

    def search_version(self, text: str, last: bool = False) -> VersionTag | None:
        """
        Search the first or last version tag occurrence in text.

        It searches for complete versions only (aka `major`, `minor` and `patch`)
        """
        candidates = (
            m for regex in self.version_regexes if len(m := list(regex.finditer(text)))
        )
        if not (matches := next(candidates, [])):
            return None

        match = matches[-1 if last else 0]

        if "version" in match.groupdict():
            return VersionTag(match.group("version"), match.group(0))

        parts = match.groupdict()
        try:
            version = f"{parts['major']}.{parts['minor']}.{parts['patch']}"
        except KeyError:
            return None

        if parts.get("prerelease"):
            version = f"{version}-{parts['prerelease']}"
        if parts.get("devrelease"):
            version = f"{version}{parts['devrelease']}"
        return VersionTag(version, match.group(0))

    def normalize_tag(
        self, version: Version | str, tag_format: str | None = None
    ) -> str:
        """
        The tag and the software version might be different.

        That's why this function exists.

        Example:
        | tag | version (PEP 0440) |
        | --- | ------- |
        | v0.9.0 | 0.9.0 |
        | ver1.0.0 | 1.0.0 |
        | ver1.0.0.a0 | 1.0.0a0 |
        """
        version = self.scheme(version) if isinstance(version, str) else version
        tag_format = tag_format or self.tag_format

        major, minor, patch = version.release
        prerelease = version.prerelease or ""

        t = Template(tag_format)
        return t.safe_substitute(
            version=version,
            major=major,
            minor=minor,
            patch=patch,
            prerelease=prerelease,
        )

    def find_tag_for(
        self, tags: Iterable[GitTag], version: Version | str
    ) -> GitTag | None:
        """Find the first matching tag for a given version."""
        version = self.scheme(version) if isinstance(version, str) else version
        possible_tags = set(self.normalize_tag(version, f) for f in self.tag_formats)
        candidates = [t for t in tags if t.name in possible_tags]
        if len(candidates) > 1:
            warnings.warn(
                UserWarning(
                    f"Multiple tags found for version {version}: {', '.join(t.name for t in candidates)}"
                )
            )
        return next(iter(candidates), None)

    @classmethod
    def from_settings(cls, settings: Settings) -> Self:
        """Extract tag rules from settings"""
        return cls(
            scheme=get_version_scheme(settings),
            tag_format=settings["tag_format"],
            legacy_tag_formats=settings["legacy_tag_formats"],
            ignored_tag_formats=settings["ignored_tag_formats"],
            merge_prereleases=settings["changelog_merge_prerelease"],
        )



from __future__ import annotations

import pytest

from commitizen.version_schemes import SemVer2, VersionProtocol
from tests.utils import VersionSchemeTestArgs


@pytest.mark.parametrize(
    "version_args, expected_version",
    [
        (
            VersionSchemeTestArgs(
                current_version="0.1.1",
                increment="PATCH",
                prerelease=None,
                prerelease_offset=0,
                devrelease=None,
            ),
            "0.1.2",
        ),
        (
            VersionSchemeTestArgs(
                current_version="0.1.1",
                increment="MINOR",
                prerelease=None,
                prerelease_offset=0,
                devrelease=None,
            ),
            "0.2.0",
        ),
        (
            VersionSchemeTestArgs(
                current_version="2.1.1",
                increment="MAJOR",
                prerelease=None,
                prerelease_offset=0,
                devrelease=None,
            ),
            "3.0.0",
        ),
        (
            VersionSchemeTestArgs(
                current_version="0.9.0",
                increment="PATCH",
                prerelease="alpha",
                prerelease_offset=0,
                devrelease=None,
            ),
            "0.9.1-alpha.0",
        ),
        (
            VersionSchemeTestArgs(
                current_version="0.9.0",
                increment="MINOR",
                prerelease="alpha",
                prerelease_offset=0,
                devrelease=None,
            ),
            "0.10.0-alpha.0",
        ),
        (
            VersionSchemeTestArgs(
                current_version="0.9.0",
                increment="MAJOR",
                prerelease="alpha",
                prerelease_offset=0,
                devrelease=None,
            ),
            "1.0.0-alpha.0",
        ),
        (
            VersionSchemeTestArgs(
                current_version="0.9.0",
                increment="MAJOR",
                prerelease="alpha",
                prerelease_offset=1,
                devrelease=None,
            ),
            "1.0.0-alpha.1",
        ),
        (
            VersionSchemeTestArgs(
                current_version="1.0.0-alpha.2",
                increment=None,
                prerelease="beta",
                prerelease_offset=0,
                devrelease=None,
            ),
            "1.0.0-beta.0",
        ),
        (
            VersionSchemeTestArgs(
                current_version="1.0.0-alpha.2",
                increment=None,
                prerelease="beta",
                prerelease_offset=1,
                devrelease=None,
            ),
            "1.0.0-beta.1",
        ),
        (
            VersionSchemeTestArgs(
                current_version="1.0.0-beta.1",
                increment=None,
                prerelease="rc",
                prerelease_offset=0,
                devrelease=None,
            ),
            "1.0.0-rc.0",
        ),
        (
            VersionSchemeTestArgs(
                current_version="1.0.0-rc.1",
                increment=None,
                prerelease="rc",
                prerelease_offset=0,
                devrelease=None,
            ),
            "1.0.0-rc.2",
        ),
        (
            VersionSchemeTestArgs(
                current_version="1.0.0-alpha.0",
                increment=None,
                prerelease="rc",
                prerelease_offset=0,
                devrelease=None,
            ),
            "1.0.0-rc.0",
        ),
        (
            VersionSchemeTestArgs(
                current_version="1.0.0-alpha.1",
                increment=None,
                prerelease="alpha",
                prerelease_offset=0,
                devrelease=None,
            ),
            "1.0.0-alpha.2",
        ),
        # weird_cases
        (
            VersionSchemeTestArgs(
                current_version="1.1",
                increment="PATCH",
                prerelease=None,
                prerelease_offset=0,
                devrelease=None,
            ),
            "1.1.1",
        ),
        (
            VersionSchemeTestArgs(
                current_version="1",
                increment="MINOR",
                prerelease=None,
                prerelease_offset=0,
                devrelease=None,
            ),
            "1.1.0",
        ),
        (
            VersionSchemeTestArgs(
                current_version="1",
                increment="MAJOR",
                prerelease=None,
                prerelease_offset=0,
                devrelease=None,
            ),
            "2.0.0",
        ),
        (
            VersionSchemeTestArgs(
                current_version="1-alpha.0",
                increment=None,
                prerelease="alpha",
                prerelease_offset=0,
                devrelease=None,
            ),
            "1.0.0-alpha.1",
        ),
        (
            VersionSchemeTestArgs(
                current_version="1-alpha.0",
                increment=None,
                prerelease="alpha",
                prerelease_offset=1,
                devrelease=None,
            ),
            "1.0.0-alpha.1",
        ),
        (
            VersionSchemeTestArgs(
                current_version="1",
                increment=None,
                prerelease="beta",
                prerelease_offset=0,
                devrelease=None,
            ),
            "1.0.0-beta.0",
        ),
        (
            VersionSchemeTestArgs(
                current_version="1",
                increment=None,
                prerelease="beta",
                prerelease_offset=1,
                devrelease=None,
            ),
            "1.0.0-beta.1",
        ),
        (
            VersionSchemeTestArgs(
                current_version="1-beta",
                increment=None,
                prerelease="beta",
                prerelease_offset=0,
                devrelease=None,
            ),
            "1.0.0-beta.1",
        ),
        (
            VersionSchemeTestArgs(
                current_version="1.0.0-alpha.1",
                increment=None,
                prerelease="alpha",
                prerelease_offset=0,
                devrelease=None,
            ),
            "1.0.0-alpha.2",
        ),
        (
            VersionSchemeTestArgs(
                current_version="1",
                increment=None,
                prerelease="rc",
                prerelease_offset=0,
                devrelease=None,
            ),
            "1.0.0-rc.0",
        ),
        (
            VersionSchemeTestArgs(
                current_version="1.0.0-rc.1+e20d7b57f3eb",
                increment="PATCH",
                prerelease=None,
                prerelease_offset=0,
                devrelease=None,
            ),
            "1.0.0",
        ),
        # simple_flow
        (
            VersionSchemeTestArgs(
                current_version="0.1.0",
                increment="PATCH",
                prerelease=None,
                prerelease_offset=0,
                devrelease=None,
            ),
            "0.1.1",
        ),
        (
            VersionSchemeTestArgs(
                current_version="0.1.0",
                increment="PATCH",
                prerelease=None,
                prerelease_offset=0,
                devrelease=1,
            ),
            "0.1.1-dev.1",
        ),
        (
            VersionSchemeTestArgs(
                current_version="0.1.1",
                increment="MINOR",
                prerelease=None,
                prerelease_offset=0,
                devrelease=None,
            ),
            "0.2.0",
        ),
        (
            VersionSchemeTestArgs(
                current_version="0.2.0",
                increment="MINOR",
                prerelease=None,
                prerelease_offset=0,
                devrelease=None,
            ),
            "0.3.0",
        ),
        (
            VersionSchemeTestArgs(
                current_version="0.2.0",
                increment="MINOR",
                prerelease=None,
                prerelease_offset=0,
                devrelease=1,
            ),
            "0.3.0-dev.1",
        ),
        (
            VersionSchemeTestArgs(
                current_version="0.3.0",
                increment="PATCH",
                prerelease=None,
                prerelease_offset=0,
                devrelease=None,
            ),
            "0.3.1",
        ),
        (
            VersionSchemeTestArgs(
                current_version="0.3.0",
                increment="PATCH",
                prerelease="alpha",
                prerelease_offset=0,
                devrelease=None,
            ),
            "0.3.1-alpha.0",
        ),
        (
            VersionSchemeTestArgs(
                current_version="0.3.1-alpha.0",
                increment=None,
                prerelease="alpha",
                prerelease_offset=0,
                devrelease=None,
            ),
            "0.3.1-alpha.1",
        ),
        (
            VersionSchemeTestArgs(
                current_version="0.3.0",
                increment="PATCH",
                prerelease="alpha",
                prerelease_offset=1,
                devrelease=None,
            ),
            "0.3.1-alpha.1",
        ),
        (
            VersionSchemeTestArgs(
                current_version="0.3.1-alpha.0",
                increment=None,
                prerelease="alpha",
                prerelease_offset=1,
                devrelease=None,
            ),
            "0.3.1-alpha.1",
        ),
        (
            VersionSchemeTestArgs(
                current_version="0.3.1-alpha.0",
                increment=None,
                prerelease=None,
                prerelease_offset=0,
                devrelease=None,
            ),
            "0.3.1",
        ),
        (
            VersionSchemeTestArgs(
                current_version="0.3.1",
                increment="PATCH",
                prerelease=None,
                prerelease_offset=0,
                devrelease=None,
            ),
            "0.3.2",
        ),
        (
            VersionSchemeTestArgs(
                current_version="0.4.2",
                increment="MAJOR",
                prerelease="alpha",
                prerelease_offset=0,
                devrelease=None,
            ),
            "1.0.0-alpha.0",
        ),
        (
            VersionSchemeTestArgs(
                current_version="1.0.0-alpha.0",
                increment=None,
                prerelease="alpha",
                prerelease_offset=0,
                devrelease=None,
            ),
            "1.0.0-alpha.1",
        ),
        (
            VersionSchemeTestArgs(
                current_version="1.0.0-alpha.1",
                increment=None,
                prerelease="alpha",
                prerelease_offset=0,
                devrelease=None,
            ),
            "1.0.0-alpha.2",
        ),
        (
            VersionSchemeTestArgs(
                current_version="1.0.0-alpha.1",
                increment=None,
                prerelease="alpha",
                prerelease_offset=0,
                devrelease=1,
            ),
            "1.0.0-alpha.2.dev.1",
        ),
        (
            VersionSchemeTestArgs(
                current_version="1.0.0-alpha.2.dev.0",
                increment=None,
                prerelease="alpha",
                prerelease_offset=0,
                devrelease=1,
            ),
            "1.0.0-alpha.3.dev.1",
        ),
        (
            VersionSchemeTestArgs(
                current_version="1.0.0-alpha.2.dev.0",
                increment=None,
                prerelease="alpha",
                prerelease_offset=0,
                devrelease=0,
            ),
            "1.0.0-alpha.3.dev.0",
        ),
        (
            VersionSchemeTestArgs(
                current_version="1.0.0-alpha.1",
                increment=None,
                prerelease="beta",
                prerelease_offset=0,
                devrelease=None,
            ),
            "1.0.0-beta.0",
        ),
        (
            VersionSchemeTestArgs(
                current_version="1.0.0-beta.0",
                increment=None,
                prerelease="beta",
                prerelease_offset=0,
                devrelease=None,
            ),
            "1.0.0-beta.1",
        ),
        (
            VersionSchemeTestArgs(
                current_version="1.0.0-beta.1",
                increment=None,
                prerelease="rc",
                prerelease_offset=0,
                devrelease=None,
            ),
            "1.0.0-rc.0",
        ),
        (
            VersionSchemeTestArgs(
                current_version="1.0.0-rc.0",
                increment=None,
                prerelease="rc",
                prerelease_offset=0,
                devrelease=None,
            ),
            "1.0.0-rc.1",
        ),
        (
            VersionSchemeTestArgs(
                current_version="1.0.0-rc.0",
                increment=None,
                prerelease="rc",
                prerelease_offset=0,
                devrelease=1,
            ),
            "1.0.0-rc.1.dev.1",
        ),
        (
            VersionSchemeTestArgs(
                current_version="1.0.0-rc.0",
                increment="PATCH",
                prerelease=None,
                prerelease_offset=0,
                devrelease=None,
            ),
            "1.0.0",
        ),
        (
            VersionSchemeTestArgs(
                current_version="1.0.0-alpha.3.dev.0",
                increment=None,
                prerelease="beta",
                prerelease_offset=0,
                devrelease=None,
            ),
            "1.0.0-beta.0",
        ),
        (
            VersionSchemeTestArgs(
                current_version="1.0.0",
                increment="PATCH",
                prerelease=None,
                prerelease_offset=0,
                devrelease=None,
            ),
            "1.0.1",
        ),
        (
            VersionSchemeTestArgs(
                current_version="1.0.1",
                increment="PATCH",
                prerelease=None,
                prerelease_offset=0,
                devrelease=None,
            ),
            "1.0.2",
        ),
        (
            VersionSchemeTestArgs(
                current_version="1.0.2",
                increment="MINOR",
                prerelease=None,
                prerelease_offset=0,
                devrelease=None,
            ),
            "1.1.0",
        ),
        (
            VersionSchemeTestArgs(
                current_version="1.1.0",
                increment="MINOR",
                prerelease=None,
                prerelease_offset=0,
                devrelease=None,
            ),
            "1.2.0",
        ),
        (
            VersionSchemeTestArgs(
                current_version="1.2.0",
                increment="PATCH",
                prerelease=None,
                prerelease_offset=0,
                devrelease=None,
            ),
            "1.2.1",
        ),
        (
            VersionSchemeTestArgs(
                current_version="1.2.1",
                increment="MAJOR",
                prerelease=None,
                prerelease_offset=0,
                devrelease=None,
            ),
            "2.0.0",
        ),
        # linear prerelease cases (never bump backwards on pre-releases)
        (
            VersionSchemeTestArgs(
                current_version="0.1.1-beta.1",
                increment=None,
                prerelease="alpha",
                prerelease_offset=0,
                devrelease=None,
            ),
            "0.1.1-beta.2",
        ),
        (
            VersionSchemeTestArgs(
                current_version="0.1.1-rc.0",
                increment=None,
                prerelease="alpha",
                prerelease_offset=0,
                devrelease=None,
            ),
            "0.1.1-rc.1",
        ),
        (
            VersionSchemeTestArgs(
                current_version="0.1.1-rc.0",
                increment=None,
                prerelease="beta",
                prerelease_offset=0,
                devrelease=None,
            ),
            "0.1.1-rc.1",
        ),
    ],
)
def test_bump_semver_version(
    version_args: VersionSchemeTestArgs, expected_version: str
):
    assert (
        str(
            SemVer2(version_args.current_version).bump(
                increment=version_args.increment,
                prerelease=version_args.prerelease,
                prerelease_offset=version_args.prerelease_offset,
                devrelease=version_args.devrelease,
            )
        )
        == expected_version
    )


@pytest.mark.parametrize(
    "version_args, expected_version",
    [
        (
            VersionSchemeTestArgs(
                current_version="4.5.0+0.1.0",
                increment="PATCH",
                prerelease=None,
                prerelease_offset=0,
                devrelease=None,
            ),
            "4.5.0+0.1.1",
        ),
        (
            VersionSchemeTestArgs(
                current_version="4.5.0+0.1.1",
                increment="MINOR",
                prerelease=None,
                prerelease_offset=0,
                devrelease=None,
            ),
            "4.5.0+0.2.0",
        ),
        (
            VersionSchemeTestArgs(
                current_version="4.5.0+0.2.0",
                increment="MAJOR",
                prerelease=None,
                prerelease_offset=0,
                devrelease=None,
            ),
            "4.5.0+1.0.0",
        ),
    ],
)
def test_bump_semver_version_local(
    version_args: VersionSchemeTestArgs, expected_version: str
):
    assert (
        str(
            SemVer2(version_args.current_version).bump(
                increment=version_args.increment,
                prerelease=version_args.prerelease,
                prerelease_offset=version_args.prerelease_offset,
                devrelease=version_args.devrelease,
                is_local_version=True,
            )
        )
        == expected_version
    )


def test_semver_scheme_property():
    version = SemVer2("0.0.1")
    assert version.scheme is SemVer2


def test_semver_implement_version_protocol():
    assert isinstance(SemVer2("0.0.1"), VersionProtocol)



from __future__ import annotations

from itertools import zip_longest
from typing import IO

from commitizen.changelog import Metadata

from .base import BaseFormat


class RestructuredText(BaseFormat):
    extension = "rst"

    def get_metadata_from_file(self, file: IO[str]) -> Metadata:
        """
        RestructuredText section titles are not one-line-based,
        they spread on 2 or 3 lines and levels are not predefined
        but determined by their occurrence order.

        It requires its own algorithm.

        For a more generic approach, you need to rely on `docutils`.
        """
        out_metadata = Metadata()
        unreleased_title_kind: str | tuple[str, str] | None = None
        is_overlined_title = False
        lines = [line.strip().lower() for line in file.readlines()]

        for index, (first, second, third) in enumerate(
            zip_longest(lines, lines[1:], lines[2:], fillvalue="")
        ):
            title: str | None = None
            kind: str | tuple[str, str] | None = None
            if _is_overlined_title(first, second, third):
                title = second
                kind = (first[0], third[0])
                is_overlined_title = True
            elif not is_overlined_title and _is_underlined_title(first, second):
                title = first
                kind = second[0]
            else:
                is_overlined_title = False

            if not title:
                continue

            if "unreleased" in title:
                unreleased_title_kind = kind
                out_metadata.unreleased_start = index
                continue

            if unreleased_title_kind and unreleased_title_kind == kind:
                out_metadata.unreleased_end = index
            # Try to find the latest release done
            if version := self.tag_rules.search_version(title):
                out_metadata.latest_version = version[0]
                out_metadata.latest_version_tag = version[1]
                out_metadata.latest_version_position = index
                break

        if (
            out_metadata.unreleased_start is not None
            and out_metadata.unreleased_end is None
        ):
            out_metadata.unreleased_end = (
                out_metadata.latest_version_position
                if out_metadata.latest_version
                else len(lines)
            )

        return out_metadata


def _is_overlined_title(first: str, second: str, third: str) -> bool:
    return (
        len(first) == len(third) >= len(second)
        and first[0] == third[0]
        and all(char == first[0] for char in first)
        and _is_underlined_title(second, third)
    )


def _is_underlined_title(first: str, second: str) -> bool:
    return (
        len(second) >= len(first)
        and not second.isalnum()
        and all(char == second[0] for char in second)
    )



import os
import subprocess
from pathlib import Path

from rich.console import Console

from commitizen.cli import data

project_root = Path(__file__).parent.parent.absolute()
images_root = project_root / Path("docs") / Path("images") / Path("cli_help")


def gen_cli_help_screenshots() -> None:
    """Generate the screenshot for help message on each cli command and save them as svg files."""
    if not os.path.exists(images_root):
        os.makedirs(images_root)
        print(f"Created {images_root}")

    help_cmds = _list_help_cmds()
    for cmd in help_cmds:
        file_name = f"{cmd.replace(' ', '_').replace('-', '_')}.svg"
        _export_cmd_as_svg(cmd, f"{images_root}/{file_name}")


def _list_help_cmds() -> list[str]:
    cmds = [f"{data['prog']} --help"] + [
        f"{data['prog']} {sub_c['name'] if isinstance(sub_c['name'], str) else sub_c['name'][0]} --help"
        for sub_c in data["subcommands"]["commands"]
    ]

    return cmds


def _export_cmd_as_svg(cmd: str, file_name: str) -> None:
    stdout = subprocess.run(cmd, shell=True, capture_output=True).stdout.decode("utf-8")
    console = Console(record=True, width=80)
    console.print(f"$ {cmd}\n{stdout}")
    console.save_svg(file_name, title="")


if __name__ == "__main__":
    gen_cli_help_screenshots()



from __future__ import annotations

import os
from pathlib import Path
from typing import TYPE_CHECKING

import pytest

if TYPE_CHECKING:
    from collections.abc import Iterator


@pytest.fixture
def chdir(tmp_path: Path) -> Iterator[Path]:
    cwd = Path()
    os.chdir(tmp_path)
    yield tmp_path
    os.chdir(cwd)



from __future__ import annotations

from typing import TYPE_CHECKING

from commitizen.providers import get_provider
from commitizen.providers.uv_provider import UvProvider

if TYPE_CHECKING:
    from pytest_regressions.file_regression import FileRegressionFixture

    from commitizen.config.base_config import BaseConfig


PYPROJECT_TOML = """
[project]
name = "test-uv"
version = "4.2.1"
description = "Add your description here"
readme = "README.md"
requires-python = ">=3.13"
dependencies = ["commitizen==4.2.1"]
"""

UV_LOCK_SIMPLIFIED = """
version = 1
revision = 1
requires-python = ">=3.13"

[[package]]
name = "commitizen"
version = "4.2.1"
source = { registry = "https://pypi.org/simple" }
dependencies = [
    { name = "argcomplete" },
    { name = "charset-normalizer" },
    { name = "colorama" },
    { name = "decli" },
    { name = "jinja2" },
    { name = "packaging" },
    { name = "pyyaml" },
    { name = "questionary" },
    { name = "termcolor" },
    { name = "tomlkit" },
]
sdist = { url = "https://files.pythonhosted.org/packages/d8/a3/77ffc9aee014cbf46c84c9f156a1ddef2d4c7cfb87d567decf2541464245/commitizen-4.2.1.tar.gz", hash = "sha256:5255416f6d6071068159f0b97605777f3e25d00927ff157b7a8d01efeda7b952", size = 50645 }
wheels = [
    { url = "https://files.pythonhosted.org/packages/57/ce/2f5d8ebe8376991b5f805e9f33d20c7f4c9ca6155bdbda761117dc41dff1/commitizen-4.2.1-py3-none-any.whl", hash = "sha256:a347889e0fe408c3b920a34130d8f35616be3ea8ac6b7b20c5b9aac19762661b", size = 72646 },
]

[[package]]
name = "decli"
version = "0.6.2"
source = { registry = "https://pypi.org/simple" }
sdist = { url = "https://files.pythonhosted.org/packages/3d/a0/a4658f93ecb589f479037b164dc13c68d108b50bf6594e54c820749f97ac/decli-0.6.2.tar.gz", hash = "sha256:36f71eb55fd0093895efb4f416ec32b7f6e00147dda448e3365cf73ceab42d6f", size = 7424 }
wheels = [
    { url = "https://files.pythonhosted.org/packages/bf/70/3ea48dc9e958d7d66c44c9944809181f1ca79aaef25703c023b5092d34ff/decli-0.6.2-py3-none-any.whl", hash = "sha256:2fc84106ce9a8f523ed501ca543bdb7e416c064917c12a59ebdc7f311a97b7ed", size = 7854 },
]

[[package]]
name = "test-uv"
version = "4.2.1"
source = { virtual = "." }
dependencies = [
    { name = "commitizen" },
]
"""


def test_uv_provider(
    config: BaseConfig, tmpdir, file_regression: FileRegressionFixture
):
    with tmpdir.as_cwd():
        pyproject_toml_file = tmpdir / UvProvider.filename
        pyproject_toml_file.write_text(PYPROJECT_TOML, encoding="utf-8")

        uv_lock_file = tmpdir / UvProvider.lock_filename
        uv_lock_file.write_text(UV_LOCK_SIMPLIFIED, encoding="utf-8")

        config.settings["version_provider"] = "uv"

        provider = get_provider(config)
        assert isinstance(provider, UvProvider)
        assert provider.get_version() == "4.2.1"

        provider.set_version("100.100.100")
        assert provider.get_version() == "100.100.100"

        updated_pyproject_toml_content = pyproject_toml_file.read_text(encoding="utf-8")
        updated_uv_lock_content = uv_lock_file.read_text(encoding="utf-8")

        for content in (updated_pyproject_toml_content, updated_uv_lock_content):
            # updated project version
            assert "100.100.100" in content
            # commitizen version which was the same as project version and should not be affected
            assert "4.2.1" in content

        file_regression.check(updated_pyproject_toml_content, extension=".toml")
        file_regression.check(updated_uv_lock_content, extension=".lock")



#!/usr/bin/env python
import shutil
import subprocess
import sys
from pathlib import Path

try:
    from commitizen.cz.utils import get_backup_file_path
except ImportError as error:
    print("could not import commitizen:")
    print(error)
    exit(1)


def prepare_commit_msg(commit_msg_file: str) -> int:
    # check if the commit message needs to be generated using commitizen
    exit_code = subprocess.run(
        [
            "cz",
            "check",
            "--commit-msg-file",
            commit_msg_file,
        ],
        capture_output=True,
    ).returncode
    if exit_code == 0:
        return 0

    backup_file = Path(get_backup_file_path())
    if backup_file.is_file():
        # confirm if commit message from backup file should be reused
        answer = input("retry with previous message? [y/N]: ")
        if answer.lower() == "y":
            shutil.copyfile(backup_file, commit_msg_file)
            return 0

    # use commitizen to generate the commit message
    exit_code = subprocess.run(
        [
            "cz",
            "commit",
            "--dry-run",
            "--write-message-to-file",
            commit_msg_file,
        ],
        stdin=sys.stdin,
        stdout=sys.stdout,
    ).returncode
    if exit_code:
        return exit_code

    # write message to backup file
    shutil.copyfile(commit_msg_file, backup_file)
    return 0


if __name__ == "__main__":
    # make hook interactive by attaching /dev/tty to stdin
    with open("/dev/tty") as tty:
        sys.stdin = tty
        exit_code = prepare_commit_msg(sys.argv[1])
        exit(exit_code)



import platform
import sys
from typing import TypedDict

from commitizen import out
from commitizen.__version__ import __version__
from commitizen.config import BaseConfig
from commitizen.exceptions import NoVersionSpecifiedError, VersionSchemeUnknown
from commitizen.providers import get_provider
from commitizen.version_schemes import get_version_scheme


class VersionArgs(TypedDict, total=False):
    commitizen: bool
    report: bool
    project: bool
    verbose: bool
    major: bool
    minor: bool


class Version:
    """Get the version of the installed commitizen or the current project.
    Precedence:
    1. report
    2. commitizen
    3. verbose, project
    """

    def __init__(self, config: BaseConfig, arguments: VersionArgs) -> None:
        self.config: BaseConfig = config
        self.arguments = arguments

    def __call__(self) -> None:
        if self.arguments.get("report"):
            out.write(f"Commitizen Version: {__version__}")
            out.write(f"Python Version: {sys.version}")
            out.write(f"Operating System: {platform.system()}")
            return

        if self.arguments.get("verbose"):
            out.write(f"Installed Commitizen Version: {__version__}")

        if not self.arguments.get("commitizen") and (
            self.arguments.get("project") or self.arguments.get("verbose")
        ):
            try:
                version = get_provider(self.config).get_version()
            except NoVersionSpecifiedError:
                out.error("No project information in this project.")
                return
            try:
                version_scheme = get_version_scheme(self.config.settings)(version)
            except VersionSchemeUnknown:
                out.error("Unknown version scheme.")
                return

            if self.arguments.get("major"):
                version = f"{version_scheme.major}"
            elif self.arguments.get("minor"):
                version = f"{version_scheme.minor}"

            out.write(
                f"Project Version: {version}"
                if self.arguments.get("verbose")
                else version
            )
            return

        if self.arguments.get("major") or self.arguments.get("minor"):
            out.error(
                "Major or minor version can only be used with --project or --verbose."
            )
            return

        # If no arguments are provided, just show the installed commitizen version
        out.write(__version__)



from __future__ import annotations

import os
from typing import TYPE_CHECKING

from tomlkit import TOMLDocument, exceptions, parse, table

from commitizen.exceptions import InvalidConfigurationError

from .base_config import BaseConfig

if TYPE_CHECKING:
    import sys
    from pathlib import Path

    # Self is Python 3.11+ but backported in typing-extensions
    if sys.version_info < (3, 11):
        from typing_extensions import Self
    else:
        from typing import Self


class TomlConfig(BaseConfig):
    def __init__(self, *, data: bytes | str, path: Path) -> None:
        super().__init__()
        self.path = path
        self._parse_setting(data)

    def init_empty_config_content(self) -> None:
        config_doc = TOMLDocument()
        if os.path.isfile(self.path):
            with open(self.path, "rb") as input_toml_file:
                config_doc = parse(input_toml_file.read())

        if config_doc.get("tool") is None:
            config_doc["tool"] = table()
        config_doc["tool"]["commitizen"] = table()  # type: ignore[index]

        with open(self.path, "wb") as output_toml_file:
            output_toml_file.write(
                config_doc.as_string().encode(self._settings["encoding"])
            )

    def set_key(self, key: str, value: object) -> Self:
        with open(self.path, "rb") as f:
            config_doc = parse(f.read())

        config_doc["tool"]["commitizen"][key] = value  # type: ignore[index]
        with open(self.path, "wb") as f:
            f.write(config_doc.as_string().encode(self._settings["encoding"]))

        return self

    def _parse_setting(self, data: bytes | str) -> None:
        """We expect to have a section in pyproject looking like

        ```
        [tool.commitizen]
        name = "cz_conventional_commits"
        ```
        """
        try:
            doc = parse(data)
        except exceptions.ParseError as e:
            raise InvalidConfigurationError(f"Failed to parse {self.path}: {e}")

        try:
            self.settings.update(doc["tool"]["commitizen"])  # type: ignore[index,typeddict-item] # TODO: fix this
        except exceptions.NonExistentKey:
            self.is_empty_config = True



"""
CC: Conventional commits
SVE: Semantic version at the end
"""

import pytest

from commitizen import bump
from commitizen.cz.conventional_commits import ConventionalCommitsCz
from commitizen.git import GitCommit

NONE_INCREMENT_CC = [
    "docs(README): motivation",
    "ci: added travis",
    "performance. Remove or disable the reimplemented linters",
    "refactor that how this line starts",
]

PATCH_INCREMENTS_CC = [
    "fix(setup.py): future is now required for every python version",
    "docs(README): motivation",
]

MINOR_INCREMENTS_CC = [
    "feat(cli): added version",
    "docs(README): motivation",
    "fix(setup.py): future is now required for every python version",
    "perf: app is much faster",
    "refactor: app is much faster",
]

MAJOR_INCREMENTS_BREAKING_CHANGE_CC = [
    "feat(cli): added version",
    "docs(README): motivation",
    "BREAKING CHANGE: `extends` key in config file is now used for extending other config files",
    "fix(setup.py): future is now required for every python version",
]

MAJOR_INCREMENTS_BREAKING_CHANGE_ALT_CC = [
    "feat(cli): added version",
    "docs(README): motivation",
    "BREAKING-CHANGE: `extends` key in config file is now used for extending other config files",
    "fix(setup.py): future is now required for every python version",
]

MAJOR_INCREMENTS_EXCLAMATION_CC = [
    "feat(cli)!: added version",
    "docs(README): motivation",
    "fix(setup.py): future is now required for every python version",
]

MAJOR_INCREMENTS_EXCLAMATION_CC_SAMPLE_2 = [
    "feat(pipeline)!: some text with breaking change"
]

MAJOR_INCREMENTS_EXCLAMATION_OTHER_TYPE_CC = [
    "chore!: drop support for Python 3.9",
    "docs(README): motivation",
    "fix(setup.py): future is now required for every python version",
]

MAJOR_INCREMENTS_EXCLAMATION_OTHER_TYPE_WITH_SCOPE_CC = [
    "chore(deps)!: drop support for Python 3.9",
    "docs(README): motivation",
    "fix(setup.py): future is now required for every python version",
]

PATCH_INCREMENTS_SVE = ["readme motivation PATCH", "fix setup.py PATCH"]

MINOR_INCREMENTS_SVE = [
    "readme motivation PATCH",
    "fix setup.py PATCH",
    "added version to cli MINOR",
]

MAJOR_INCREMENTS_SVE = [
    "readme motivation PATCH",
    "fix setup.py PATCH",
    "added version to cli MINOR",
    "extends key is used for other config files MAJOR",
]

semantic_version_pattern = r"(MAJOR|MINOR|PATCH)"
semantic_version_map = {"MAJOR": "MAJOR", "MINOR": "MINOR", "PATCH": "PATCH"}


@pytest.mark.parametrize(
    "messages, expected_type",
    (
        (PATCH_INCREMENTS_CC, "PATCH"),
        (MINOR_INCREMENTS_CC, "MINOR"),
        (MAJOR_INCREMENTS_BREAKING_CHANGE_CC, "MAJOR"),
        (MAJOR_INCREMENTS_BREAKING_CHANGE_ALT_CC, "MAJOR"),
        (MAJOR_INCREMENTS_EXCLAMATION_OTHER_TYPE_CC, "MAJOR"),
        (MAJOR_INCREMENTS_EXCLAMATION_OTHER_TYPE_WITH_SCOPE_CC, "MAJOR"),
        (MAJOR_INCREMENTS_EXCLAMATION_CC, "MAJOR"),
        (MAJOR_INCREMENTS_EXCLAMATION_CC_SAMPLE_2, "MAJOR"),
        (NONE_INCREMENT_CC, None),
    ),
)
def test_find_increment(messages, expected_type):
    commits = [GitCommit(rev="test", title=message) for message in messages]
    increment_type = bump.find_increment(
        commits,
        regex=ConventionalCommitsCz.bump_pattern,
        increments_map=ConventionalCommitsCz.bump_map,
    )
    assert increment_type == expected_type


@pytest.mark.parametrize(
    "messages, expected_type",
    (
        (PATCH_INCREMENTS_SVE, "PATCH"),
        (MINOR_INCREMENTS_SVE, "MINOR"),
        (MAJOR_INCREMENTS_SVE, "MAJOR"),
    ),
)
def test_find_increment_sve(messages, expected_type):
    commits = [GitCommit(rev="test", title=message) for message in messages]
    increment_type = bump.find_increment(
        commits, regex=semantic_version_pattern, increments_map=semantic_version_map
    )
    assert increment_type == expected_type



from __future__ import annotations

from commitizen.exceptions import NoVersionSpecifiedError
from commitizen.providers.base_provider import VersionProvider


class CommitizenProvider(VersionProvider):
    """
    Default version provider: Fetch and set version in commitizen config.
    """

    def get_version(self) -> str:
        if ret := self.config.settings["version"]:
            return ret
        raise NoVersionSpecifiedError()

    def set_version(self, version: str) -> None:
        self.config.set_key("version", version)



from __future__ import annotations

from typing import TYPE_CHECKING

from commitizen.providers.commitizen_provider import CommitizenProvider

if TYPE_CHECKING:
    from pytest_mock import MockerFixture

    from commitizen.config.base_config import BaseConfig


def test_commitizen_provider(config: BaseConfig, mocker: MockerFixture):
    config.settings["version"] = "42"
    mock = mocker.patch.object(config, "set_key")

    provider = CommitizenProvider(config)
    assert provider.get_version() == "42"

    provider.set_version("43.1")
    mock.assert_called_once_with("version", "43.1")



#!/usr/bin/env python

try:
    from commitizen.cz.utils import get_backup_file_path
except ImportError as error:
    print(f"could not import commitizen:\n{error}")
    exit(1)


def post_commit() -> None:
    backup_file_path = get_backup_file_path()

    # remove backup file if it exists
    if backup_file_path.is_file():
        backup_file_path.unlink()


if __name__ == "__main__":
    post_commit()
    exit(0)



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from commitizen.changelog import Metadata
from commitizen.changelog_formats.asciidoc import AsciiDoc

if TYPE_CHECKING:
    from pathlib import Path

    from commitizen.config.base_config import BaseConfig

CHANGELOG_A = """
= Changelog

All notable changes to this project will be documented in this file.

The format is based on https://keepachangelog.com/en/1.0.0/[Keep a Changelog],
and this project adheres to https://semver.org/spec/v2.0.0.html[Semantic Versioning].

== [Unreleased]
* Start using "changelog" over "change log" since it's the common usage.

== [1.0.0] - 2017-06-20
=== Added
* New visual identity by https://github.com/tylerfortune8[@tylerfortune8].
* Version navigation.
""".strip()

EXPECTED_A = Metadata(
    latest_version="1.0.0",
    latest_version_position=10,
    unreleased_end=10,
    unreleased_start=7,
)


CHANGELOG_B = """
== [Unreleased]
* Start using "changelog" over "change log" since it's the common usage.

== 1.2.0
""".strip()

EXPECTED_B = Metadata(
    latest_version="1.2.0",
    latest_version_position=3,
    unreleased_end=3,
    unreleased_start=0,
)


CHANGELOG_C = """
= Unreleased

== v1.0.0
"""
EXPECTED_C = Metadata(
    latest_version="1.0.0",
    latest_version_tag="v1.0.0",
    latest_version_position=3,
    unreleased_end=3,
    unreleased_start=1,
)

CHANGELOG_D = """
== Unreleased
* Start using "changelog" over "change log" since it's the common usage.
"""

EXPECTED_D = Metadata(
    latest_version=None,
    latest_version_position=None,
    unreleased_end=2,
    unreleased_start=1,
)

CHANGELOG_E = """
= Changelog

All notable changes to this project will be documented in this file.

The format is based on https://keepachangelog.com/en/1.0.0/[Keep a Changelog],
and this project adheres to https://semver.org/spec/v2.0.0.html[Semantic Versioning].

== [Unreleased]
* Start using "changelog" over "change log" since it's the common usage.

== [{tag_formatted_version}] - 2017-06-20
=== Added
* New visual identity by https://github.com/tylerfortune8[@tylerfortune8].
* Version navigation.
""".strip()

EXPECTED_E = Metadata(
    latest_version="1.0.0",
    latest_version_position=10,
    unreleased_end=10,
    unreleased_start=7,
)


@pytest.fixture
def format(config: BaseConfig) -> AsciiDoc:
    return AsciiDoc(config)


@pytest.fixture
def format_with_tags(config: BaseConfig, request) -> AsciiDoc:
    config.settings["tag_format"] = request.param
    config.settings["legacy_tag_formats"] = ["legacy-${version}"]
    return AsciiDoc(config)


VERSIONS_EXAMPLES = [
    ("== [1.0.0] - 2017-06-20", ("1.0.0", "1.0.0")),
    (
        "= https://github.com/angular/angular/compare/10.0.0-next.2...10.0.0-next.3[10.0.0-next.3] (2020-04-22)",
        ("10.0.0-next.3", "10.0.0-next.3"),
    ),
    ("=== 0.19.1 (Jan 7, 2020)", ("0.19.1", "0.19.1")),
    ("== 1.0.0", ("1.0.0", "1.0.0")),
    ("== v1.0.0", ("1.0.0", "v1.0.0")),
    ("== v1.0.0 - (2012-24-32)", ("1.0.0", "v1.0.0")),
    ("= version 2020.03.24", ("2020.03.24", "2020.03.24")),
    ("== [Unreleased]", None),
    ("All notable changes to this project will be documented in this file.", None),
    ("= Changelog", None),
    ("=== Bug Fixes", None),
]


@pytest.mark.parametrize("line_from_changelog,output_version", VERSIONS_EXAMPLES)
def test_changelog_detect_version(
    line_from_changelog: str, output_version: tuple[str, str] | None, format: AsciiDoc
):
    version = format.parse_version_from_title(line_from_changelog)
    assert version == output_version


TITLES_EXAMPLES = [
    ("== [1.0.0] - 2017-06-20", 2),
    ("== [Unreleased]", 2),
    ("= Unreleased", 1),
]


@pytest.mark.parametrize("line_from_changelog,output_title", TITLES_EXAMPLES)
def test_parse_title_type_of_line(
    line_from_changelog: str, output_title: str, format: AsciiDoc
):
    title = format.parse_title_level(line_from_changelog)
    assert title == output_title


@pytest.mark.parametrize(
    "content, expected",
    (
        pytest.param(CHANGELOG_A, EXPECTED_A, id="A"),
        pytest.param(CHANGELOG_B, EXPECTED_B, id="B"),
        pytest.param(CHANGELOG_C, EXPECTED_C, id="C"),
        pytest.param(CHANGELOG_D, EXPECTED_D, id="D"),
    ),
)
def test_get_metadata(
    tmp_path: Path, format: AsciiDoc, content: str, expected: Metadata
):
    changelog = tmp_path / format.default_changelog_file
    changelog.write_text(content)

    assert format.get_metadata(str(changelog)) == expected


@pytest.mark.parametrize(
    "format_with_tags, tag_string, expected, ",
    (
        pytest.param("${version}-example", "1.0.0-example", "1.0.0"),
        pytest.param("${version}example", "1.0.0example", "1.0.0"),
        pytest.param("example${version}", "example1.0.0", "1.0.0"),
        pytest.param("example-${version}", "example-1.0.0", "1.0.0"),
        pytest.param("example-${major}-${minor}-${patch}", "example-1-0-0", "1.0.0"),
        pytest.param("example-${major}-${minor}", "example-1-0-0", None),
        pytest.param(
            "${major}-${minor}-${patch}-${prerelease}-example",
            "1-0-0-rc1-example",
            "1.0.0-rc1",
        ),
        pytest.param(
            "${major}-${minor}-${patch}-${prerelease}${devrelease}-example",
            "1-0-0-a1.dev1-example",
            "1.0.0-a1.dev1",
        ),
        pytest.param("new-${version}", "legacy-1.0.0", "1.0.0"),
    ),
    indirect=["format_with_tags"],
)
def test_get_metadata_custom_tag_format(
    tmp_path: Path, format_with_tags: AsciiDoc, tag_string: str, expected: Metadata
):
    content = CHANGELOG_E.format(tag_formatted_version=tag_string)
    changelog = tmp_path / format_with_tags.default_changelog_file
    changelog.write_text(content)
    assert format_with_tags.get_metadata(str(changelog)).latest_version == expected



from __future__ import annotations

from abc import ABCMeta, abstractmethod
from collections.abc import Iterable, Mapping
from typing import TYPE_CHECKING, Any, NamedTuple, Protocol

from jinja2 import BaseLoader, PackageLoader
from prompt_toolkit.styles import Style

from commitizen.exceptions import CommitMessageLengthExceededError

if TYPE_CHECKING:
    import re
    from collections.abc import Callable, Iterable, Mapping

    from commitizen import git
    from commitizen.config.base_config import BaseConfig
    from commitizen.question import CzQuestion


class MessageBuilderHook(Protocol):
    def __call__(
        self, message: dict[str, Any], commit: git.GitCommit
    ) -> dict[str, Any] | Iterable[dict[str, Any]] | None: ...


class ChangelogReleaseHook(Protocol):
    def __call__(
        self, release: dict[str, Any], tag: git.GitTag | None
    ) -> dict[str, Any]: ...


class ValidationResult(NamedTuple):
    is_valid: bool
    errors: list


class BaseCommitizen(metaclass=ABCMeta):
    bump_pattern: str | None = None
    bump_map: dict[str, str] | None = None
    bump_map_major_version_zero: dict[str, str] | None = None
    default_style_config: list[tuple[str, str]] = [
        ("qmark", "fg:#ff9d00 bold"),
        ("question", "bold"),
        ("answer", "fg:#ff9d00 bold"),
        ("pointer", "fg:#ff9d00 bold"),
        ("highlighted", "fg:#ff9d00 bold"),
        ("selected", "fg:#cc5454"),
        ("separator", "fg:#cc5454"),
        ("instruction", ""),
        ("text", ""),
        ("disabled", "fg:#858585 italic"),
    ]

    # The whole subject will be parsed as a message by default
    # This allows supporting changelog for any rule system.
    # It can be modified per rule
    commit_parser: str | None = r"(?P<message>.*)"
    changelog_pattern: str | None = r".*"
    change_type_map: dict[str, str] | None = None
    change_type_order: list[str] | None = None

    # Executed per message parsed by the commitizen
    changelog_message_builder_hook: MessageBuilderHook | None = None

    # Executed only at the end of the changelog generation
    changelog_hook: Callable[[str, str | None], str] | None = None

    # Executed for each release in the changelog
    changelog_release_hook: ChangelogReleaseHook | None = None

    # Plugins can override templates and provide extra template data
    template_loader: BaseLoader = PackageLoader("commitizen", "templates")
    template_extras: dict[str, Any] = {}

    def __init__(self, config: BaseConfig) -> None:
        self.config = config
        if not self.config.settings.get("style"):
            self.config.settings.update({"style": BaseCommitizen.default_style_config})

    @abstractmethod
    def questions(self) -> list[CzQuestion]:
        """Questions regarding the commit message."""

    @abstractmethod
    def message(self, answers: Mapping[str, Any]) -> str:
        """Format your git message."""

    @property
    def style(self) -> Style:
        return Style(
            [
                *BaseCommitizen.default_style_config,
                *self.config.settings["style"],
            ]
        )

    @abstractmethod
    def example(self) -> str:
        """Example of the commit message."""

    @abstractmethod
    def schema(self) -> str:
        """Schema definition of the commit message."""

    @abstractmethod
    def schema_pattern(self) -> str:
        """Regex matching the schema used for message validation."""

    @abstractmethod
    def info(self) -> str:
        """Information about the standardized commit message."""

    def validate_commit_message(
        self,
        *,
        commit_msg: str,
        pattern: re.Pattern[str],
        allow_abort: bool,
        allowed_prefixes: list[str],
        max_msg_length: int | None,
        commit_hash: str,
    ) -> ValidationResult:
        """Validate commit message against the pattern."""
        if not commit_msg:
            return ValidationResult(
                allow_abort, [] if allow_abort else ["commit message is empty"]
            )

        if any(map(commit_msg.startswith, allowed_prefixes)):
            return ValidationResult(True, [])

        if max_msg_length is not None:
            msg_len = len(commit_msg.partition("\n")[0].strip())
            if msg_len > max_msg_length:
                # TODO: capitalize the first letter of the error message for consistency in v5
                raise CommitMessageLengthExceededError(
                    f"commit validation: failed!\n"
                    f"commit message length exceeds the limit.\n"
                    f'commit "{commit_hash}": "{commit_msg}"\n'
                    f"message length limit: {max_msg_length} (actual: {msg_len})"
                )

        return ValidationResult(
            bool(pattern.match(commit_msg)),
            [f"pattern: {pattern.pattern}"],
        )

    def format_exception_message(
        self, invalid_commits: list[tuple[git.GitCommit, list]]
    ) -> str:
        """Format commit errors."""
        displayed_msgs_content = "\n".join(
            [
                f'commit "{commit.rev}": "{commit.message}\n"' + "\n".join(errors)
                for commit, errors in invalid_commits
            ]
        )
        # TODO: capitalize the first letter of the error message for consistency in v5
        return (
            "commit validation: failed!\n"
            "please enter a commit message in the commitizen format.\n"
            f"{displayed_msgs_content}"
        )



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from commitizen import defaults
from commitizen.changelog_formats import (
    KNOWN_CHANGELOG_FORMATS,
    ChangelogFormat,
    _guess_changelog_format,
    get_changelog_format,
)
from commitizen.exceptions import ChangelogFormatUnknown

if TYPE_CHECKING:
    from commitizen.config.base_config import BaseConfig


@pytest.mark.parametrize("format", KNOWN_CHANGELOG_FORMATS.values())
def test_guess_format(format: type[ChangelogFormat]):
    assert _guess_changelog_format(f"CHANGELOG.{format.extension}") is format
    for ext in format.alternative_extensions:
        assert _guess_changelog_format(f"CHANGELOG.{ext}") is format


@pytest.mark.parametrize("filename", ("CHANGELOG", "NEWS", "file.unknown", None))
def test_guess_format_unknown(filename: str):
    assert _guess_changelog_format(filename) is None


@pytest.mark.parametrize(
    "name, expected",
    [
        pytest.param(name, format, id=name)
        for name, format in KNOWN_CHANGELOG_FORMATS.items()
    ],
)
def test_get_format(config: BaseConfig, name: str, expected: type[ChangelogFormat]):
    config.settings["changelog_format"] = name
    assert isinstance(get_changelog_format(config), expected)


@pytest.mark.parametrize("filename", (None, ""))
def test_get_format_empty_filename(config: BaseConfig, filename: str | None):
    config.settings["changelog_format"] = defaults.CHANGELOG_FORMAT
    assert isinstance(
        get_changelog_format(config, filename),
        KNOWN_CHANGELOG_FORMATS[defaults.CHANGELOG_FORMAT],
    )


@pytest.mark.parametrize("filename", (None, ""))
def test_get_format_empty_filename_no_setting(config: BaseConfig, filename: str | None):
    config.settings["changelog_format"] = None
    with pytest.raises(ChangelogFormatUnknown):
        get_changelog_format(config, filename)


@pytest.mark.parametrize("filename", ("extensionless", "file.unknown"))
def test_get_format_unknown(config: BaseConfig, filename: str | None):
    with pytest.raises(ChangelogFormatUnknown):
        get_changelog_format(config, filename)



from __future__ import annotations

import re
from typing import TYPE_CHECKING

from .base import BaseFormat

if TYPE_CHECKING:
    from commitizen.tags import VersionTag


class AsciiDoc(BaseFormat):
    extension = "adoc"

    RE_TITLE = re.compile(r"^(?P<level>=+) (?P<title>.*)$")

    def parse_version_from_title(self, line: str) -> VersionTag | None:
        m = self.RE_TITLE.match(line)
        if not m:
            return None
        # Capture last match as AsciiDoc use postfixed URL labels
        return self.tag_rules.search_version(m.group("title"), last=True)

    def parse_title_level(self, line: str) -> int | None:
        m = self.RE_TITLE.match(line)
        if not m:
            return None
        return len(m.group("level"))



from __future__ import annotations

import os
import re
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Any
from unittest.mock import Mock

import pytest
from jinja2 import FileSystemLoader

from commitizen import changelog, git
from commitizen.commands.changelog import Changelog
from commitizen.config import BaseConfig
from commitizen.cz.conventional_commits.conventional_commits import (
    ConventionalCommitsCz,
)
from commitizen.exceptions import InvalidConfigurationError
from commitizen.version_schemes import Pep440

if TYPE_CHECKING:
    from commitizen.changelog_formats import ChangelogFormat

COMMITS_DATA: list[dict[str, Any]] = [
    {
        "rev": "141ee441c9c9da0809c554103a558eb17c30ed17",
        "parents": ["6c4948501031b7d6405b54b21d3d635827f9421b"],
        "title": "bump: version 1.1.1 → 1.2.0",
        "body": "",
        "author": "Commitizen",
        "author_email": "author@cz.dev",
    },
    {
        "rev": "6c4948501031b7d6405b54b21d3d635827f9421b",
        "parents": ["ddd220ad515502200fe2dde443614c1075d26238"],
        "title": "docs: how to create custom bumps",
        "body": "",
        "author": "Commitizen",
        "author_email": "author@cz.dev",
    },
    {
        "rev": "ddd220ad515502200fe2dde443614c1075d26238",
        "parents": ["ad17acff2e3a2e141cbc3c6efd7705e4e6de9bfc"],
        "title": "feat: custom cz plugins now support bumping version",
        "body": "",
        "author": "Commitizen",
        "author_email": "author@cz.dev",
    },
    {
        "rev": "ad17acff2e3a2e141cbc3c6efd7705e4e6de9bfc",
        "parents": ["56c8a8da84e42b526bcbe130bd194306f7c7e813"],
        "title": "docs: added bump gif",
        "body": "",
        "author": "Commitizen",
        "author_email": "author@cz.dev",
    },
    {
        "rev": "56c8a8da84e42b526bcbe130bd194306f7c7e813",
        "parents": ["74c6134b1b2e6bb8b07ed53410faabe99b204f36"],
        "title": "bump: version 1.1.0 → 1.1.1",
        "body": "",
        "author": "Commitizen",
        "author_email": "author@cz.dev",
    },
    {
        "rev": "74c6134b1b2e6bb8b07ed53410faabe99b204f36",
        "parents": ["cbc7b5f22c4e74deff4bc92d14e19bd93524711e"],
        "title": "refactor: changed stdout statements",
        "body": "",
        "author": "Commitizen",
        "author_email": "author@cz.dev",
    },
    {
        "rev": "cbc7b5f22c4e74deff4bc92d14e19bd93524711e",
        "parents": ["1ba46f2a63cb9d6e7472eaece21528c8cd28b118"],
        "title": "fix(bump): commit message now fits better with semver",
        "body": "",
        "author": "Commitizen",
        "author_email": "author@cz.dev",
    },
    {
        "rev": "1ba46f2a63cb9d6e7472eaece21528c8cd28b118",
        "parents": ["c35dbffd1bb98bb0b3d1593797e79d1c3366af8f"],
        "title": "fix: conventional commit 'breaking change' in body instead of title",
        "body": "closes #16",
        "author": "Commitizen",
        "author_email": "author@cz.dev",
    },
    {
        "rev": "c35dbffd1bb98bb0b3d1593797e79d1c3366af8f",
        "parents": ["25313397a4ac3dc5b5c986017bee2a614399509d"],
        "title": "refactor(schema): command logic removed from commitizen base",
        "body": "",
        "author": "Commitizen",
        "author_email": "author@cz.dev",
    },
    {
        "rev": "25313397a4ac3dc5b5c986017bee2a614399509d",
        "parents": ["d2f13ac41b4e48995b3b619d931c82451886e6ff"],
        "title": "refactor(info): command logic removed from commitizen base",
        "body": "",
        "author": "Commitizen",
        "author_email": "author@cz.dev",
    },
    {
        "rev": "d2f13ac41b4e48995b3b619d931c82451886e6ff",
        "parents": ["d839e317e5b26671b010584ad8cc6bf362400fa1"],
        "title": "refactor(example): command logic removed from commitizen base",
        "body": "",
        "author": "Commitizen",
        "author_email": "author@cz.dev",
    },
    {
        "rev": "d839e317e5b26671b010584ad8cc6bf362400fa1",
        "parents": ["12d0e65beda969f7983c444ceedc2a01584f4e08"],
        "title": "refactor(commit): moved most of the commit logic to the commit command",
        "body": "",
        "author": "Commitizen",
        "author_email": "author@cz.dev",
    },
    {
        "rev": "12d0e65beda969f7983c444ceedc2a01584f4e08",
        "parents": ["fb4c85abe51c228e50773e424cbd885a8b6c610d"],
        "title": "docs(README): updated documentation url)",
        "body": "",
        "author": "Commitizen",
        "author_email": "author@cz.dev",
    },
    {
        "rev": "fb4c85abe51c228e50773e424cbd885a8b6c610d",
        "parents": ["17efb44d2cd16f6621413691a543e467c7d2dda6"],
        "title": "docs: mkdocs documentation",
        "body": "",
        "author": "Commitizen",
        "author_email": "author@cz.dev",
    },
    {
        "rev": "17efb44d2cd16f6621413691a543e467c7d2dda6",
        "parents": ["6012d9eecfce8163d75c8fff179788e9ad5347da"],
        "title": "Bump version 1.0.0 → 1.1.0",
        "body": "",
        "author": "Commitizen",
        "author_email": "author@cz.dev",
    },
    {
        "rev": "6012d9eecfce8163d75c8fff179788e9ad5347da",
        "parents": ["0c7fb0ca0168864dfc55d83c210da57771a18319"],
        "title": "test: fixed issues with conf",
        "body": "",
        "author": "Commitizen",
        "author_email": "author@cz.dev",
    },
    {
        "rev": "0c7fb0ca0168864dfc55d83c210da57771a18319",
        "parents": ["cb1dd2019d522644da5bdc2594dd6dee17122d7f"],
        "title": "docs(README): some new information about bump",
        "body": "",
        "author": "Commitizen",
        "author_email": "author@cz.dev",
    },
    {
        "rev": "cb1dd2019d522644da5bdc2594dd6dee17122d7f",
        "parents": ["9c7450f85df6bf6be508e79abf00855a30c3c73c"],
        "title": "feat: new working bump command",
        "body": "",
        "author": "Commitizen",
        "author_email": "author@cz.dev",
    },
    {
        "rev": "9c7450f85df6bf6be508e79abf00855a30c3c73c",
        "parents": ["9f3af3772baab167e3fd8775d37f041440184251"],
        "title": "feat: create version tag",
        "body": "",
        "author": "Commitizen",
        "author_email": "author@cz.dev",
    },
    {
        "rev": "9f3af3772baab167e3fd8775d37f041440184251",
        "parents": ["b0d6a3defbfde14e676e7eb34946409297d0221b"],
        "title": "docs: added new changelog",
        "body": "",
        "author": "Commitizen",
        "author_email": "author@cz.dev",
    },
    {
        "rev": "b0d6a3defbfde14e676e7eb34946409297d0221b",
        "parents": ["d630d07d912e420f0880551f3ac94e933f9d3beb"],
        "title": "feat: update given files with new version",
        "body": "",
        "author": "Commitizen",
        "author_email": "author@cz.dev",
    },
    {
        "rev": "d630d07d912e420f0880551f3ac94e933f9d3beb",
        "parents": ["1792b8980c58787906dbe6836f93f31971b1ec2d"],
        "title": "fix: removed all from commit",
        "body": "",
        "author": "Commitizen",
        "author_email": "author@cz.dev",
    },
    {
        "rev": "1792b8980c58787906dbe6836f93f31971b1ec2d",
        "parents": ["52def1ea3555185ba4b936b463311949907e31ec"],
        "title": "feat(config): new set key, used to set version to cfg",
        "body": "",
        "author": "Commitizen",
        "author_email": "author@cz.dev",
    },
    {
        "rev": "52def1ea3555185ba4b936b463311949907e31ec",
        "parents": ["3127e05077288a5e2b62893345590bf1096141b7"],
        "title": "feat: support for pyproject.toml",
        "body": "",
        "author": "Commitizen",
        "author_email": "author@cz.dev",
    },
    {
        "rev": "3127e05077288a5e2b62893345590bf1096141b7",
        "parents": ["fd480ed90a80a6ffa540549408403d5b60d0e90c"],
        "title": "feat: first semantic version bump implementation",
        "body": "",
        "author": "Commitizen",
        "author_email": "author@cz.dev",
    },
    {
        "rev": "fd480ed90a80a6ffa540549408403d5b60d0e90c",
        "parents": ["e4840a059731c0bf488381ffc77e989e85dd81ad"],
        "title": "fix: fix config file not working",
        "body": "",
        "author": "Commitizen",
        "author_email": "author@cz.dev",
    },
    {
        "rev": "e4840a059731c0bf488381ffc77e989e85dd81ad",
        "parents": ["aa44a92d68014d0da98965c0c2cb8c07957d4362"],
        "title": "refactor: added commands folder, better integration with decli",
        "body": "",
        "author": "Commitizen",
        "author_email": "author@cz.dev",
    },
    {
        "rev": "aa44a92d68014d0da98965c0c2cb8c07957d4362",
        "parents": ["58bb709765380dbd46b74ce6e8978515764eb955"],
        "title": "Bump version: 1.0.0b2 → 1.0.0",
        "body": "",
        "author": "Commitizen",
        "author_email": "author@cz.dev",
    },
    {
        "rev": "58bb709765380dbd46b74ce6e8978515764eb955",
        "parents": ["97afb0bb48e72b6feca793091a8a23c706693257"],
        "title": "docs(README): new badges",
        "body": "",
        "author": "Commitizen",
        "author_email": "author@cz.dev",
    },
    {
        "rev": "97afb0bb48e72b6feca793091a8a23c706693257",
        "parents": [
            "9cecb9224aa7fa68d4afeac37eba2a25770ef251",
            "e004a90b81ea5b374f118759bce5951202d03d69",
        ],
        "title": "Merge pull request #10 from Woile/feat/decli",
        "body": "Feat/decli",
        "author": "Commitizen",
        "author_email": "author@cz.dev",
    },
    {
        "rev": "9cecb9224aa7fa68d4afeac37eba2a25770ef251",
        "parents": ["f5781d1a2954d71c14ade2a6a1a95b91310b2577"],
        "title": "style: black to files",
        "body": "",
        "author": "Commitizen",
        "author_email": "author@cz.dev",
    },
    {
        "rev": "f5781d1a2954d71c14ade2a6a1a95b91310b2577",
        "parents": ["80105fb3c6d45369bc0cbf787bd329fba603864c"],
        "title": "ci: added travis",
        "body": "",
        "author": "Commitizen",
        "author_email": "author@cz.dev",
    },
    {
        "rev": "80105fb3c6d45369bc0cbf787bd329fba603864c",
        "parents": ["a96008496ffefb6b1dd9b251cb479eac6a0487f7"],
        "title": "refactor: removed delegator, added decli and many tests",
        "body": "BREAKING CHANGE: API is stable",
        "author": "Commitizen",
        "author_email": "author@cz.dev",
    },
    {
        "rev": "a96008496ffefb6b1dd9b251cb479eac6a0487f7",
        "parents": ["aab33d13110f26604fb786878856ec0b9e5fc32b"],
        "title": "docs: updated test command",
        "body": "",
        "author": "Commitizen",
        "author_email": "author@cz.dev",
    },
    {
        "rev": "aab33d13110f26604fb786878856ec0b9e5fc32b",
        "parents": ["b73791563d2f218806786090fb49ef70faa51a3a"],
        "title": "Bump version: 1.0.0b1 → 1.0.0b2",
        "body": "",
        "author": "Commitizen",
        "author_email": "author@cz.dev",
    },
    {
        "rev": "b73791563d2f218806786090fb49ef70faa51a3a",
        "parents": ["7aa06a454fb717408b3657faa590731fb4ab3719"],
        "title": "docs(README): updated to reflect current state",
        "body": "",
        "author": "Commitizen",
        "author_email": "author@cz.dev",
    },
    {
        "rev": "7aa06a454fb717408b3657faa590731fb4ab3719",
        "parents": [
            "7c7e96b723c2aaa1aec3a52561f680adf0b60e97",
            "9589a65880016996cff156b920472b9d28d771ca",
        ],
        "title": "Merge pull request #9 from Woile/dev",
        "body": "feat: py3 only, tests and conventional commits 1.0",
        "author": "Commitizen",
        "author_email": "author@cz.dev",
    },
    {
        "rev": "7c7e96b723c2aaa1aec3a52561f680adf0b60e97",
        "parents": ["ed830019581c83ba633bfd734720e6758eca6061"],
        "title": "Bump version: 0.9.11 → 1.0.0b1",
        "body": "",
        "author": "Commitizen",
        "author_email": "author@cz.dev",
    },
    {
        "rev": "ed830019581c83ba633bfd734720e6758eca6061",
        "parents": ["c52eca6f74f844ab3ffbde61d98ef96071e132b7"],
        "title": "feat: py3 only, tests and conventional commits 1.0",
        "body": "more tests\npyproject instead of Pipfile\nquestionary instead of whaaaaat (promptkit 2.0.0 support)",
        "author": "Commitizen",
        "author_email": "author@cz.dev",
    },
    {
        "rev": "c52eca6f74f844ab3ffbde61d98ef96071e132b7",
        "parents": ["0326652b2657083929507ee66d4d1a0899e861ba"],
        "title": "Bump version: 0.9.10 → 0.9.11",
        "body": "",
        "author": "Commitizen",
        "author_email": "author@cz.dev",
    },
    {
        "rev": "0326652b2657083929507ee66d4d1a0899e861ba",
        "parents": ["b3f89892222340150e32631ae6b7aab65230036f"],
        "title": "fix(config): load config reads in order without failing if there is no commitizen section",
        "body": "Closes #8",
        "author": "Commitizen",
        "author_email": "author@cz.dev",
    },
    {
        "rev": "b3f89892222340150e32631ae6b7aab65230036f",
        "parents": ["5e837bf8ef0735193597372cd2d85e31a8f715b9"],
        "title": "Bump version: 0.9.9 → 0.9.10",
        "body": "",
        "author": "Commitizen",
        "author_email": "author@cz.dev",
    },
    {
        "rev": "5e837bf8ef0735193597372cd2d85e31a8f715b9",
        "parents": ["684e0259cc95c7c5e94854608cd3dcebbd53219e"],
        "title": "fix: parse scope (this is my punishment for not having tests)",
        "body": "",
        "author": "Commitizen",
        "author_email": "author@cz.dev",
    },
    {
        "rev": "684e0259cc95c7c5e94854608cd3dcebbd53219e",
        "parents": ["ca38eac6ff09870851b5c76a6ff0a2a8e5ecda15"],
        "title": "Bump version: 0.9.8 → 0.9.9",
        "body": "",
        "author": "Commitizen",
        "author_email": "author@cz.dev",
    },
    {
        "rev": "ca38eac6ff09870851b5c76a6ff0a2a8e5ecda15",
        "parents": ["64168f18d4628718c49689ee16430549e96c5d4b"],
        "title": "fix: parse scope empty",
        "body": "",
        "author": "Commitizen",
        "author_email": "author@cz.dev",
    },
    {
        "rev": "64168f18d4628718c49689ee16430549e96c5d4b",
        "parents": ["9d4def716ef235a1fa5ae61614366423fbc8256f"],
        "title": "Bump version: 0.9.7 → 0.9.8",
        "body": "",
        "author": "Commitizen",
        "author_email": "author@cz.dev",
    },
    {
        "rev": "9d4def716ef235a1fa5ae61614366423fbc8256f",
        "parents": ["33b0bf1a0a4dc60aac45ed47476d2e5473add09e"],
        "title": "fix(scope): parse correctly again",
        "body": "",
        "author": "Commitizen",
        "author_email": "author@cz.dev",
    },
    {
        "rev": "33b0bf1a0a4dc60aac45ed47476d2e5473add09e",
        "parents": ["696885e891ec35775daeb5fec3ba2ab92c2629e1"],
        "title": "Bump version: 0.9.6 → 0.9.7",
        "body": "",
        "author": "Commitizen",
        "author_email": "author@cz.dev",
    },
    {
        "rev": "696885e891ec35775daeb5fec3ba2ab92c2629e1",
        "parents": ["bef4a86761a3bda309c962bae5d22ce9b57119e4"],
        "title": "fix(scope): parse correctly",
        "body": "",
        "author": "Commitizen",
        "author_email": "author@cz.dev",
    },
    {
        "rev": "bef4a86761a3bda309c962bae5d22ce9b57119e4",
        "parents": ["72472efb80f08ee3fd844660afa012c8cb256e4b"],
        "title": "Bump version: 0.9.5 → 0.9.6",
        "body": "",
        "author": "Commitizen",
        "author_email": "author@cz.dev",
    },
    {
        "rev": "72472efb80f08ee3fd844660afa012c8cb256e4b",
        "parents": ["b5561ce0ab3b56bb87712c8f90bcf37cf2474f1b"],
        "title": "refactor(conventionalCommit): moved filters to questions instead of message",
        "body": "",
        "author": "Commitizen",
        "author_email": "author@cz.dev",
    },
    {
        "rev": "b5561ce0ab3b56bb87712c8f90bcf37cf2474f1b",
        "parents": ["3e31714dc737029d96898f412e4ecd2be1bcd0ce"],
        "title": "fix(manifest): included missing files",
        "body": "",
        "author": "Commitizen",
        "author_email": "author@cz.dev",
    },
    {
        "rev": "3e31714dc737029d96898f412e4ecd2be1bcd0ce",
        "parents": ["9df721e06595fdd216884c36a28770438b4f4a39"],
        "title": "Bump version: 0.9.4 → 0.9.5",
        "body": "",
        "author": "Commitizen",
        "author_email": "author@cz.dev",
    },
    {
        "rev": "9df721e06595fdd216884c36a28770438b4f4a39",
        "parents": ["0cf6ada372470c8d09e6c9e68ebf94bbd5a1656f"],
        "title": "fix(config): home path for python versions between 3.0 and 3.5",
        "body": "",
        "author": "Commitizen",
        "author_email": "author@cz.dev",
    },
    {
        "rev": "0cf6ada372470c8d09e6c9e68ebf94bbd5a1656f",
        "parents": ["973c6b3e100f6f69a3fe48bd8ee55c135b96c318"],
        "title": "Bump version: 0.9.3 → 0.9.4",
        "body": "",
        "author": "Commitizen",
        "author_email": "author@cz.dev",
    },
    {
        "rev": "973c6b3e100f6f69a3fe48bd8ee55c135b96c318",
        "parents": ["dacc86159b260ee98eb5f57941c99ba731a01399"],
        "title": "feat(cli): added version",
        "body": "",
        "author": "Commitizen",
        "author_email": "author@cz.dev",
    },
    {
        "rev": "dacc86159b260ee98eb5f57941c99ba731a01399",
        "parents": ["4368f3c3cbfd4a1ced339212230d854bc5bab496"],
        "title": "Bump version: 0.9.2 → 0.9.3",
        "body": "",
        "author": "Commitizen",
        "author_email": "author@cz.dev",
    },
    {
        "rev": "4368f3c3cbfd4a1ced339212230d854bc5bab496",
        "parents": ["da94133288727d35dae9b91866a25045038f2d38"],
        "title": "feat(committer): conventional commit is a bit more intelligent now",
        "body": "",
        "author": "Commitizen",
        "author_email": "author@cz.dev",
    },
    {
        "rev": "da94133288727d35dae9b91866a25045038f2d38",
        "parents": ["1541f54503d2e1cf39bd777c0ca5ab5eb78772ba"],
        "title": "docs(README): motivation",
        "body": "",
        "author": "Commitizen",
        "author_email": "author@cz.dev",
    },
    {
        "rev": "1541f54503d2e1cf39bd777c0ca5ab5eb78772ba",
        "parents": ["ddc855a637b7879108308b8dbd85a0fd27c7e0e7"],
        "title": "Bump version: 0.9.1 → 0.9.2",
        "body": "",
        "author": "Commitizen",
        "author_email": "author@cz.dev",
    },
    {
        "rev": "ddc855a637b7879108308b8dbd85a0fd27c7e0e7",
        "parents": ["46e9032e18a819e466618c7a014bcb0e9981af9e"],
        "title": "refactor: renamed conventional_changelog to conventional_commits, not backward compatible",
        "body": "",
        "author": "Commitizen",
        "author_email": "author@cz.dev",
    },
    {
        "rev": "46e9032e18a819e466618c7a014bcb0e9981af9e",
        "parents": ["0fef73cd7dc77a25b82e197e7c1d3144a58c1350"],
        "title": "Bump version: 0.9.0 → 0.9.1",
        "body": "",
        "author": "Commitizen",
        "author_email": "author@cz.dev",
    },
    {
        "rev": "0fef73cd7dc77a25b82e197e7c1d3144a58c1350",
        "parents": [],
        "title": "fix(setup.py): future is now required for every python version",
        "body": "",
        "author": "Commitizen",
        "author_email": "author@cz.dev",
    },
]


TAGS = [
    ("v1.2.0", "141ee441c9c9da0809c554103a558eb17c30ed17", "2019-04-19"),
    ("v1.1.1", "56c8a8da84e42b526bcbe130bd194306f7c7e813", "2019-04-18"),
    ("v1.1.0", "17efb44d2cd16f6621413691a543e467c7d2dda6", "2019-04-14"),
    ("v1.0.0", "aa44a92d68014d0da98965c0c2cb8c07957d4362", "2019-03-01"),
    ("1.0.0b2", "aab33d13110f26604fb786878856ec0b9e5fc32b", "2019-01-18"),
    ("v1.0.0b1", "7c7e96b723c2aaa1aec3a52561f680adf0b60e97", "2019-01-17"),
    ("v0.9.11", "c52eca6f74f844ab3ffbde61d98ef96071e132b7", "2018-12-17"),
    ("v0.9.10", "b3f89892222340150e32631ae6b7aab65230036f", "2018-09-22"),
    ("v0.9.9", "684e0259cc95c7c5e94854608cd3dcebbd53219e", "2018-09-22"),
    ("v0.9.8", "64168f18d4628718c49689ee16430549e96c5d4b", "2018-09-22"),
    ("v0.9.7", "33b0bf1a0a4dc60aac45ed47476d2e5473add09e", "2018-09-22"),
    ("v0.9.6", "bef4a86761a3bda309c962bae5d22ce9b57119e4", "2018-09-19"),
    ("v0.9.5", "3e31714dc737029d96898f412e4ecd2be1bcd0ce", "2018-08-24"),
    ("v0.9.4", "0cf6ada372470c8d09e6c9e68ebf94bbd5a1656f", "2018-08-02"),
    ("v0.9.3", "dacc86159b260ee98eb5f57941c99ba731a01399", "2018-07-28"),
    ("v0.9.2", "1541f54503d2e1cf39bd777c0ca5ab5eb78772ba", "2017-11-11"),
    ("v0.9.1", "46e9032e18a819e466618c7a014bcb0e9981af9e", "2017-11-11"),
]


@pytest.fixture
def gitcommits() -> list[git.GitCommit]:
    return [
        git.GitCommit(
            commit["rev"],
            commit["title"],
            commit["body"],
            commit["author"],
            commit["author_email"],
            commit["parents"],
        )
        for commit in COMMITS_DATA
    ]


@pytest.fixture
def tags() -> list[git.GitTag]:
    return [git.GitTag(*tag) for tag in TAGS]


@pytest.fixture
def changelog_content(data_dir: Path) -> str:
    changelog = data_dir / "CHANGELOG_FOR_TEST.md"
    with changelog.open(encoding="utf-8") as f:
        return f.read()


def test_get_commit_tag_is_a_version(gitcommits, tags):
    commit = gitcommits[0]
    tag = git.GitTag(*TAGS[0])
    current_key = changelog.get_commit_tag(commit, tags)
    assert current_key == tag


def test_get_commit_tag_is_None(gitcommits, tags):
    commit = gitcommits[1]
    current_key = changelog.get_commit_tag(commit, tags)
    assert current_key is None


@pytest.mark.parametrize("test_input", TAGS)
def test_valid_tag_included_in_changelog(test_input):
    tag = git.GitTag(*test_input)
    rules = changelog.TagRules()
    assert rules.include_in_changelog(tag)


def test_invalid_tag_included_in_changelog():
    tag = git.GitTag("not_a_version", "rev", "date")
    rules = changelog.TagRules()
    assert not rules.include_in_changelog(tag)


COMMITS_TREE = (
    {
        "version": "v1.2.0",
        "date": "2019-04-19",
        "changes": {
            "feat": [
                {
                    "scope": None,
                    "breaking": None,
                    "message": "custom cz plugins now support bumping version",
                }
            ]
        },
    },
    {
        "version": "v1.1.1",
        "date": "2019-04-18",
        "changes": {
            "refactor": [
                {
                    "scope": None,
                    "breaking": None,
                    "message": "changed stdout statements",
                },
                {
                    "scope": "schema",
                    "breaking": None,
                    "message": "command logic removed from commitizen base",
                },
                {
                    "scope": "info",
                    "breaking": None,
                    "message": "command logic removed from commitizen base",
                },
                {
                    "scope": "example",
                    "breaking": None,
                    "message": "command logic removed from commitizen base",
                },
                {
                    "scope": "commit",
                    "breaking": None,
                    "message": "moved most of the commit logic to the commit command",
                },
            ],
            "fix": [
                {
                    "scope": "bump",
                    "breaking": None,
                    "message": "commit message now fits better with semver",
                },
                {
                    "scope": None,
                    "breaking": None,
                    "message": "conventional commit 'breaking change' in body instead of title",
                },
            ],
        },
    },
    {
        "version": "v1.1.0",
        "date": "2019-04-14",
        "changes": {
            "feat": [
                {
                    "scope": None,
                    "breaking": None,
                    "message": "new working bump command",
                },
                {"scope": None, "breaking": None, "message": "create version tag"},
                {
                    "scope": None,
                    "breaking": None,
                    "message": "update given files with new version",
                },
                {
                    "scope": "config",
                    "breaking": None,
                    "message": "new set key, used to set version to cfg",
                },
                {
                    "scope": None,
                    "breaking": None,
                    "message": "support for pyproject.toml",
                },
                {
                    "scope": None,
                    "breaking": None,
                    "message": "first semantic version bump implementation",
                },
            ],
            "fix": [
                {
                    "scope": None,
                    "breaking": None,
                    "message": "removed all from commit",
                },
                {
                    "scope": None,
                    "breaking": None,
                    "message": "fix config file not working",
                },
            ],
            "refactor": [
                {
                    "scope": None,
                    "breaking": None,
                    "message": "added commands folder, better integration with decli",
                }
            ],
        },
    },
    {
        "version": "v1.0.0",
        "date": "2019-03-01",
        "changes": {
            "refactor": [
                {
                    "scope": None,
                    "breaking": None,
                    "message": "removed delegator, added decli and many tests",
                }
            ],
            "BREAKING CHANGE": [
                {"scope": None, "breaking": None, "message": "API is stable"}
            ],
        },
    },
    {"version": "1.0.0b2", "date": "2019-01-18", "changes": {}},
    {
        "version": "v1.0.0b1",
        "date": "2019-01-17",
        "changes": {
            "feat": [
                {
                    "scope": None,
                    "breaking": None,
                    "message": "py3 only, tests and conventional commits 1.0",
                }
            ]
        },
    },
    {
        "version": "v0.9.11",
        "date": "2018-12-17",
        "changes": {
            "fix": [
                {
                    "scope": "config",
                    "breaking": None,
                    "message": "load config reads in order without failing if there is no commitizen section",
                }
            ]
        },
    },
    {
        "version": "v0.9.10",
        "date": "2018-09-22",
        "changes": {
            "fix": [
                {
                    "scope": None,
                    "breaking": None,
                    "message": "parse scope (this is my punishment for not having tests)",
                }
            ]
        },
    },
    {
        "version": "v0.9.9",
        "date": "2018-09-22",
        "changes": {
            "fix": [{"scope": None, "breaking": None, "message": "parse scope empty"}]
        },
    },
    {
        "version": "v0.9.8",
        "date": "2018-09-22",
        "changes": {
            "fix": [
                {
                    "scope": "scope",
                    "breaking": None,
                    "message": "parse correctly again",
                }
            ]
        },
    },
    {
        "version": "v0.9.7",
        "date": "2018-09-22",
        "changes": {
            "fix": [{"scope": "scope", "breaking": None, "message": "parse correctly"}]
        },
    },
    {
        "version": "v0.9.6",
        "date": "2018-09-19",
        "changes": {
            "refactor": [
                {
                    "scope": "conventionalCommit",
                    "breaking": None,
                    "message": "moved filters to questions instead of message",
                }
            ],
            "fix": [
                {
                    "scope": "manifest",
                    "breaking": None,
                    "message": "included missing files",
                }
            ],
        },
    },
    {
        "version": "v0.9.5",
        "date": "2018-08-24",
        "changes": {
            "fix": [
                {
                    "scope": "config",
                    "breaking": None,
                    "message": "home path for python versions between 3.0 and 3.5",
                }
            ]
        },
    },
    {
        "version": "v0.9.4",
        "date": "2018-08-02",
        "changes": {
            "feat": [{"scope": "cli", "breaking": None, "message": "added version"}]
        },
    },
    {
        "version": "v0.9.3",
        "date": "2018-07-28",
        "changes": {
            "feat": [
                {
                    "scope": "committer",
                    "breaking": None,
                    "message": "conventional commit is a bit more intelligent now",
                }
            ]
        },
    },
    {
        "version": "v0.9.2",
        "date": "2017-11-11",
        "changes": {
            "refactor": [
                {
                    "scope": None,
                    "breaking": None,
                    "message": "renamed conventional_changelog to conventional_commits, not backward compatible",
                }
            ]
        },
    },
    {
        "version": "v0.9.1",
        "date": "2017-11-11",
        "changes": {
            "fix": [
                {
                    "scope": "setup.py",
                    "breaking": None,
                    "message": "future is now required for every python version",
                }
            ]
        },
    },
)

COMMITS_TREE_AFTER_MERGED_PRERELEASES = (
    {
        "version": "v1.2.0",
        "date": "2019-04-19",
        "changes": {
            "feat": [
                {
                    "scope": None,
                    "breaking": None,
                    "message": "custom cz plugins now support bumping version",
                }
            ]
        },
    },
    {
        "version": "v1.1.1",
        "date": "2019-04-18",
        "changes": {
            "refactor": [
                {
                    "scope": None,
                    "breaking": None,
                    "message": "changed stdout statements",
                },
                {
                    "scope": "schema",
                    "breaking": None,
                    "message": "command logic removed from commitizen base",
                },
                {
                    "scope": "info",
                    "breaking": None,
                    "message": "command logic removed from commitizen base",
                },
                {
                    "scope": "example",
                    "breaking": None,
                    "message": "command logic removed from commitizen base",
                },
                {
                    "scope": "commit",
                    "breaking": None,
                    "message": "moved most of the commit logic to the commit command",
                },
            ],
            "fix": [
                {
                    "scope": "bump",
                    "breaking": None,
                    "message": "commit message now fits better with semver",
                },
                {
                    "scope": None,
                    "breaking": None,
                    "message": "conventional commit 'breaking change' in body instead of title",
                },
            ],
        },
    },
    {
        "version": "v1.1.0",
        "date": "2019-04-14",
        "changes": {
            "feat": [
                {
                    "scope": None,
                    "breaking": None,
                    "message": "new working bump command",
                },
                {"scope": None, "breaking": None, "message": "create version tag"},
                {
                    "scope": None,
                    "breaking": None,
                    "message": "update given files with new version",
                },
                {
                    "scope": "config",
                    "breaking": None,
                    "message": "new set key, used to set version to cfg",
                },
                {
                    "scope": None,
                    "breaking": None,
                    "message": "support for pyproject.toml",
                },
                {
                    "scope": None,
                    "breaking": None,
                    "message": "first semantic version bump implementation",
                },
            ],
            "fix": [
                {
                    "scope": None,
                    "breaking": None,
                    "message": "removed all from commit",
                },
                {
                    "scope": None,
                    "breaking": None,
                    "message": "fix config file not working",
                },
            ],
            "refactor": [
                {
                    "scope": None,
                    "breaking": None,
                    "message": "added commands folder, better integration with decli",
                }
            ],
        },
    },
    {
        "version": "v1.0.0",
        "date": "2019-03-01",
        "changes": {
            "refactor": [
                {
                    "scope": None,
                    "breaking": None,
                    "message": "removed delegator, added decli and many tests",
                }
            ],
            "feat": [
                {
                    "scope": None,
                    "breaking": None,
                    "message": "py3 only, tests and conventional commits 1.0",
                }
            ],
            "BREAKING CHANGE": [
                {"scope": None, "breaking": None, "message": "API is stable"}
            ],
        },
    },
    {
        "version": "v0.9.11",
        "date": "2018-12-17",
        "changes": {
            "fix": [
                {
                    "scope": "config",
                    "breaking": None,
                    "message": "load config reads in order without failing if there is no commitizen section",
                }
            ]
        },
    },
    {
        "version": "v0.9.10",
        "date": "2018-09-22",
        "changes": {
            "fix": [
                {
                    "scope": None,
                    "breaking": None,
                    "message": "parse scope (this is my punishment for not having tests)",
                }
            ]
        },
    },
    {
        "version": "v0.9.9",
        "date": "2018-09-22",
        "changes": {
            "fix": [{"scope": None, "breaking": None, "message": "parse scope empty"}]
        },
    },
    {
        "version": "v0.9.8",
        "date": "2018-09-22",
        "changes": {
            "fix": [
                {
                    "scope": "scope",
                    "breaking": None,
                    "message": "parse correctly again",
                }
            ]
        },
    },
    {
        "version": "v0.9.7",
        "date": "2018-09-22",
        "changes": {
            "fix": [{"scope": "scope", "breaking": None, "message": "parse correctly"}]
        },
    },
    {
        "version": "v0.9.6",
        "date": "2018-09-19",
        "changes": {
            "refactor": [
                {
                    "scope": "conventionalCommit",
                    "breaking": None,
                    "message": "moved filters to questions instead of message",
                }
            ],
            "fix": [
                {
                    "scope": "manifest",
                    "breaking": None,
                    "message": "included missing files",
                }
            ],
        },
    },
    {
        "version": "v0.9.5",
        "date": "2018-08-24",
        "changes": {
            "fix": [
                {
                    "scope": "config",
                    "breaking": None,
                    "message": "home path for python versions between 3.0 and 3.5",
                }
            ]
        },
    },
    {
        "version": "v0.9.4",
        "date": "2018-08-02",
        "changes": {
            "feat": [{"scope": "cli", "breaking": None, "message": "added version"}]
        },
    },
    {
        "version": "v0.9.3",
        "date": "2018-07-28",
        "changes": {
            "feat": [
                {
                    "scope": "committer",
                    "breaking": None,
                    "message": "conventional commit is a bit more intelligent now",
                }
            ]
        },
    },
    {
        "version": "v0.9.2",
        "date": "2017-11-11",
        "changes": {
            "refactor": [
                {
                    "scope": None,
                    "breaking": None,
                    "message": "renamed conventional_changelog to conventional_commits, not backward compatible",
                }
            ]
        },
    },
    {
        "version": "v0.9.1",
        "date": "2017-11-11",
        "changes": {
            "fix": [
                {
                    "scope": "setup.py",
                    "breaking": None,
                    "message": "future is now required for every python version",
                }
            ]
        },
    },
)


@pytest.mark.parametrize("merge_prereleases", (True, False))
def test_generate_tree_from_commits(gitcommits, tags, merge_prereleases):
    parser = ConventionalCommitsCz.commit_parser
    changelog_pattern = ConventionalCommitsCz.bump_pattern
    rules = changelog.TagRules(
        merge_prereleases=merge_prereleases,
    )
    tree = changelog.generate_tree_from_commits(
        gitcommits, tags, parser, changelog_pattern, rules=rules
    )
    expected = (
        COMMITS_TREE_AFTER_MERGED_PRERELEASES if merge_prereleases else COMMITS_TREE
    )

    for release, expected_release in zip(tree, expected):
        assert release["version"] == expected_release["version"]
        assert release["date"] == expected_release["date"]
        assert release["changes"].keys() == expected_release["changes"].keys()
        for change_type in release["changes"]:
            changes = release["changes"][change_type]
            expected_changes = expected_release["changes"][change_type]
            for change, expected_change in zip(changes, expected_changes):
                assert change["scope"] == expected_change["scope"]
                assert change["breaking"] == expected_change["breaking"]
                assert change["message"] == expected_change["message"]
                assert change["author"] == "Commitizen"
                assert change["author_email"] in "author@cz.dev"
                assert "sha1" in change
                assert "parents" in change


def test_generate_tree_from_commits_with_no_commits(tags):
    gitcommits = []
    parser = ConventionalCommitsCz.commit_parser
    changelog_pattern = ConventionalCommitsCz.bump_pattern
    tree = changelog.generate_tree_from_commits(
        gitcommits, tags, parser, changelog_pattern
    )

    assert tuple(tree) == ({"changes": {}, "date": "", "version": "Unreleased"},)


@pytest.mark.parametrize(
    "change_type_order, expected_reordering",
    (
        ([], {}),
        (
            ["BREAKING CHANGE", "refactor"],
            {
                "1.1.0": {
                    "original": ["feat", "fix", "refactor"],
                    "sorted": ["refactor", "feat", "fix"],
                },
                "1.0.0": {
                    "original": ["refactor", "BREAKING CHANGE"],
                    "sorted": ["BREAKING CHANGE", "refactor"],
                },
            },
        ),
    ),
)
def test_generate_ordered_changelog_tree(change_type_order, expected_reordering):
    tree = changelog.generate_ordered_changelog_tree(COMMITS_TREE, change_type_order)

    for index, entry in enumerate(tuple(tree)):
        version = entry["version"]
        if version in expected_reordering:
            # Verify that all keys are present
            assert [*entry.keys()] == [*COMMITS_TREE[index].keys()]
            # Verify that the reorder only impacted the returned dict and not the original
            expected = expected_reordering[version]
            assert [*entry["changes"].keys()] == expected["sorted"]
            assert [*COMMITS_TREE[index]["changes"].keys()] == expected["original"]
        else:
            assert [*entry["changes"].keys()] == [*entry["changes"].keys()]


def test_generate_ordered_changelog_tree_raises():
    change_type_order = ["BREAKING CHANGE", "feat", "refactor", "feat"]
    with pytest.raises(InvalidConfigurationError) as excinfo:
        list(changelog.generate_ordered_changelog_tree(COMMITS_TREE, change_type_order))

    assert "Change types contain duplicated types" in str(excinfo)


def test_render_changelog(
    gitcommits, tags, changelog_content, any_changelog_format: ChangelogFormat
):
    parser = ConventionalCommitsCz.commit_parser
    changelog_pattern = ConventionalCommitsCz.changelog_pattern
    loader = ConventionalCommitsCz.template_loader
    template = any_changelog_format.template
    tree = changelog.generate_tree_from_commits(
        gitcommits, tags, parser, changelog_pattern
    )
    result = changelog.render_changelog(tree, loader, template)
    assert result == changelog_content


def test_render_changelog_from_default_plugin_values(
    gitcommits, tags, changelog_content, any_changelog_format: ChangelogFormat
):
    parser = ConventionalCommitsCz.commit_parser
    changelog_pattern = ConventionalCommitsCz.changelog_pattern
    loader = ConventionalCommitsCz.template_loader
    template = any_changelog_format.template
    tree = changelog.generate_tree_from_commits(
        gitcommits, tags, parser, changelog_pattern
    )
    result = changelog.render_changelog(tree, loader, template)
    assert result == changelog_content


def test_render_changelog_override_loader(gitcommits, tags, tmp_path: Path):
    loader = FileSystemLoader(tmp_path)
    template = "tpl.j2"
    tpl = "loader overridden"
    (tmp_path / template).write_text(tpl)
    parser = ConventionalCommitsCz.commit_parser
    changelog_pattern = ConventionalCommitsCz.changelog_pattern
    tree = changelog.generate_tree_from_commits(
        gitcommits, tags, parser, changelog_pattern
    )
    result = changelog.render_changelog(tree, loader, template)
    assert result == tpl


def test_render_changelog_override_template_from_cwd(
    gitcommits, tags, chdir: Path, any_changelog_format: ChangelogFormat
):
    tpl = "overridden from cwd"
    template = any_changelog_format.template
    (chdir / template).write_text(tpl)
    parser = ConventionalCommitsCz.commit_parser
    changelog_pattern = ConventionalCommitsCz.changelog_pattern
    loader = ConventionalCommitsCz.template_loader
    tree = changelog.generate_tree_from_commits(
        gitcommits, tags, parser, changelog_pattern
    )
    result = changelog.render_changelog(tree, loader, template)
    assert result == tpl


def test_render_changelog_override_template_from_cwd_with_custom_name(
    gitcommits, tags, chdir: Path
):
    tpl = "template overridden from cwd"
    tpl_name = "tpl.j2"
    (chdir / tpl_name).write_text(tpl)
    parser = ConventionalCommitsCz.commit_parser
    changelog_pattern = ConventionalCommitsCz.changelog_pattern
    loader = ConventionalCommitsCz.template_loader
    tree = changelog.generate_tree_from_commits(
        gitcommits, tags, parser, changelog_pattern
    )
    result = changelog.render_changelog(tree, loader, tpl_name)
    assert result == tpl


def test_render_changelog_override_loader_and_template(
    gitcommits, tags, tmp_path: Path
):
    loader = FileSystemLoader(tmp_path)
    tpl = "loader and template overridden"
    tpl_name = "tpl.j2"
    (tmp_path / tpl_name).write_text(tpl)
    parser = ConventionalCommitsCz.commit_parser
    changelog_pattern = ConventionalCommitsCz.bump_pattern
    tree = changelog.generate_tree_from_commits(
        gitcommits, tags, parser, changelog_pattern
    )
    result = changelog.render_changelog(tree, loader, tpl_name)
    assert result == tpl


def test_render_changelog_support_arbitrary_kwargs(gitcommits, tags, tmp_path: Path):
    loader = FileSystemLoader(tmp_path)
    tpl_name = "tpl.j2"
    (tmp_path / tpl_name).write_text("{{ key }}")
    parser = ConventionalCommitsCz.commit_parser
    changelog_pattern = ConventionalCommitsCz.changelog_pattern
    tree = changelog.generate_tree_from_commits(
        gitcommits, tags, parser, changelog_pattern
    )
    result = changelog.render_changelog(tree, loader, tpl_name, key="value")
    assert result == "value"


def test_render_changelog_unreleased(gitcommits, any_changelog_format: ChangelogFormat):
    some_commits = gitcommits[:7]
    parser = ConventionalCommitsCz.commit_parser
    changelog_pattern = ConventionalCommitsCz.changelog_pattern
    loader = ConventionalCommitsCz.template_loader
    template = any_changelog_format.template
    tree = changelog.generate_tree_from_commits(
        some_commits, [], parser, changelog_pattern
    )
    result = changelog.render_changelog(tree, loader, template)
    assert "Unreleased" in result


def test_render_changelog_tag_and_unreleased(
    gitcommits, tags, any_changelog_format: ChangelogFormat
):
    some_commits = gitcommits[:7]
    single_tag = [
        tag for tag in tags if tag.rev == "56c8a8da84e42b526bcbe130bd194306f7c7e813"
    ]

    parser = ConventionalCommitsCz.commit_parser
    changelog_pattern = ConventionalCommitsCz.changelog_pattern
    loader = ConventionalCommitsCz.template_loader
    template = any_changelog_format.template
    tree = changelog.generate_tree_from_commits(
        some_commits, single_tag, parser, changelog_pattern
    )
    result = changelog.render_changelog(tree, loader, template)

    assert "Unreleased" in result
    assert "## v1.1.1" in result


def test_render_changelog_with_change_type(
    gitcommits, tags, any_changelog_format: ChangelogFormat
):
    new_title = ":some-emoji: feature"
    change_type_map = {"feat": new_title}
    parser = ConventionalCommitsCz.commit_parser
    changelog_pattern = ConventionalCommitsCz.changelog_pattern
    loader = ConventionalCommitsCz.template_loader
    template = any_changelog_format.template
    tree = changelog.generate_tree_from_commits(
        gitcommits, tags, parser, changelog_pattern, change_type_map=change_type_map
    )
    result = changelog.render_changelog(tree, loader, template)
    assert new_title in result


def test_render_changelog_with_changelog_message_builder_hook(
    gitcommits, tags, any_changelog_format: ChangelogFormat
):
    def changelog_message_builder_hook(message: dict, commit: git.GitCommit) -> dict:
        message["message"] = (
            f"{message['message']} [link](github.com/232323232) {commit.author} {commit.author_email}"
        )
        return message

    parser = ConventionalCommitsCz.commit_parser
    changelog_pattern = ConventionalCommitsCz.changelog_pattern
    loader = ConventionalCommitsCz.template_loader
    template = any_changelog_format.template
    tree = changelog.generate_tree_from_commits(
        gitcommits,
        tags,
        parser,
        changelog_pattern,
        changelog_message_builder_hook=changelog_message_builder_hook,
    )
    result = changelog.render_changelog(tree, loader, template)

    assert "[link](github.com/232323232) Commitizen author@cz.dev" in result


def test_changelog_message_builder_hook_can_remove_commits(
    gitcommits, tags, any_changelog_format: ChangelogFormat
):
    def changelog_message_builder_hook(message: dict, commit: git.GitCommit):
        return None

    parser = ConventionalCommitsCz.commit_parser
    changelog_pattern = ConventionalCommitsCz.changelog_pattern
    loader = ConventionalCommitsCz.template_loader
    template = any_changelog_format.template
    tree = changelog.generate_tree_from_commits(
        gitcommits,
        tags,
        parser,
        changelog_pattern,
        changelog_message_builder_hook=changelog_message_builder_hook,
    )
    result = changelog.render_changelog(tree, loader, template)

    RE_HEADER = re.compile(r"^## v?\d+\.\d+\.\d+(\w)* \(\d{4}-\d{2}-\d{2}\)$")
    # Rendered changelog should be empty, only containing version headers
    for no, line in enumerate(result.splitlines()):
        if line := line.strip():
            assert RE_HEADER.match(line), f"Line {no} should not be there: {line}"


def test_render_changelog_with_changelog_message_builder_hook_multiple_entries(
    gitcommits, tags, any_changelog_format: ChangelogFormat
):
    def changelog_message_builder_hook(message: dict, commit: git.GitCommit):
        messages = [message.copy(), message.copy(), message.copy()]
        for idx, msg in enumerate(messages):
            msg["message"] = "Message #{idx}"
        return messages

    parser = ConventionalCommitsCz.commit_parser
    changelog_pattern = ConventionalCommitsCz.changelog_pattern
    loader = ConventionalCommitsCz.template_loader
    template = any_changelog_format.template
    tree = changelog.generate_tree_from_commits(
        gitcommits,
        tags,
        parser,
        changelog_pattern,
        changelog_message_builder_hook=changelog_message_builder_hook,
    )
    result = changelog.render_changelog(tree, loader, template)

    for idx in range(3):
        assert "Message #{idx}" in result


def test_changelog_message_builder_hook_can_access_and_modify_change_type(
    gitcommits, tags, any_changelog_format: ChangelogFormat
):
    def changelog_message_builder_hook(message: dict, commit: git.GitCommit):
        assert "change_type" in message
        message["change_type"] = "overridden"
        return message

    parser = ConventionalCommitsCz.commit_parser
    changelog_pattern = ConventionalCommitsCz.changelog_pattern
    loader = ConventionalCommitsCz.template_loader
    template = any_changelog_format.template
    tree = changelog.generate_tree_from_commits(
        gitcommits,
        tags,
        parser,
        changelog_pattern,
        changelog_message_builder_hook=changelog_message_builder_hook,
    )
    result = changelog.render_changelog(tree, loader, template)

    RE_HEADER = re.compile(r"^### (?P<type>.+)$")
    # There should be only "overridden" change type headers
    for no, line in enumerate(result.splitlines()):
        if (line := line.strip()) and (match := RE_HEADER.match(line)):
            change_type = match.group("type")
            assert change_type == "overridden", (
                f"Line {no}: type {change_type} should have been overridden"
            )


def test_render_changelog_with_changelog_release_hook(
    gitcommits, tags, any_changelog_format: ChangelogFormat
):
    def changelog_release_hook(release: dict, tag: git.GitTag | None) -> dict:
        release["extra"] = "whatever"
        return release

    parser = ConventionalCommitsCz.commit_parser
    changelog_pattern = ConventionalCommitsCz.changelog_pattern
    tree = changelog.generate_tree_from_commits(
        gitcommits,
        tags,
        parser,
        changelog_pattern,
        changelog_release_hook=changelog_release_hook,
    )
    for release in tree:
        assert release["extra"] == "whatever"


def test_get_smart_tag_range_returns_an_extra_for_a_range(tags):
    start, end = (
        tags[0],
        tags[2],
    )  # len here is 3, but we expect one more tag as designed
    res = changelog.get_smart_tag_range(tags, start.name, end.name)
    assert 4 == len(res)


def test_get_smart_tag_range_returns_an_extra_for_a_single_tag(tags):
    start = tags[0]  # len here is 1, but we expect one more tag as designed
    res = changelog.get_smart_tag_range(tags, start.name)
    assert 2 == len(res)


def test_get_next_tag_name_after_version(tags):
    # Test finding next tag after a version
    next_tag_name = changelog.get_next_tag_name_after_version(tags, "v1.2.0")
    assert next_tag_name == "v1.1.1"

    next_tag_name = changelog.get_next_tag_name_after_version(tags, "v1.1.0")
    assert next_tag_name == "v1.0.0"

    # Test finding last tag when given version is last
    last_tag_name = changelog.get_next_tag_name_after_version(tags, "v0.9.1")
    assert last_tag_name is None

    # Test error when version not found
    with pytest.raises(changelog.NoCommitsFoundError) as exc_info:
        changelog.get_next_tag_name_after_version(tags, "nonexistent")
    assert "Could not find a valid revision range" in str(exc_info.value)


@dataclass
class TagDef:
    name: str
    is_version: bool
    is_legacy: bool
    is_ignored: bool


TAGS_PARAMS = (
    pytest.param(TagDef("1.2.3", True, False, False), id="version"),
    # We test with `v-` prefix as `v` prefix is a special case kept for backward compatibility
    pytest.param(TagDef("v-1.2.3", False, True, False), id="v-prefix"),
    pytest.param(TagDef("project-1.2.3", False, True, False), id="project-prefix"),
    pytest.param(TagDef("ignored", False, False, True), id="ignored"),
    pytest.param(TagDef("unknown", False, False, False), id="unknown"),
)


@pytest.mark.parametrize("tag", TAGS_PARAMS)
def test_tag_rules_tag_format_only(tag: TagDef):
    rules = changelog.TagRules(Pep440, "$version")
    assert rules.is_version_tag(tag.name) is tag.is_version


@pytest.mark.parametrize("tag", TAGS_PARAMS)
def test_tag_rules_with_legacy_tags(tag: TagDef):
    rules = changelog.TagRules(
        scheme=Pep440,
        tag_format="$version",
        legacy_tag_formats=["v-$version", "project-${version}"],
    )
    assert rules.is_version_tag(tag.name) is tag.is_version or tag.is_legacy


@pytest.mark.parametrize("tag", TAGS_PARAMS)
def test_tag_rules_with_ignored_tags(tag: TagDef):
    rules = changelog.TagRules(
        scheme=Pep440, tag_format="$version", ignored_tag_formats=["ignored"]
    )
    assert rules.is_ignored_tag(tag.name) is tag.is_ignored


def test_tags_rules_get_version_tags(capsys: pytest.CaptureFixture):
    tags = [
        git.GitTag("v1.1.0", "17efb44d2cd16f6621413691a543e467c7d2dda6", "2019-04-14"),
        git.GitTag("v1.0.0", "aa44a92d68014d0da98965c0c2cb8c07957d4362", "2019-03-01"),
        git.GitTag("1.0.0b2", "aab33d13110f26604fb786878856ec0b9e5fc32b", "2019-01-18"),
        git.GitTag(
            "project-not-a-version",
            "7c7e96b723c2aaa1aec3a52561f680adf0b60e97",
            "2019-01-17",
        ),
        git.GitTag(
            "not-a-version", "c52eca6f74f844ab3ffbde61d98ef96071e132b7", "2018-12-17"
        ),
        git.GitTag(
            "star-something", "c52eca6f74f844ab3ffbde61d98fe96071e132b2", "2018-11-12"
        ),
        git.GitTag("known", "b3f89892222340150e32631ae6b7aab65230036f", "2018-09-22"),
        git.GitTag(
            "ignored-0.9.3", "684e0259cc95c7c5e94854608cd3dcebbd53219e", "2018-09-22"
        ),
        git.GitTag(
            "project-0.9.3", "dacc86159b260ee98eb5f57941c99ba731a01399", "2018-07-28"
        ),
        git.GitTag(
            "anything-0.9", "5141f54503d2e1cf39bd666c0ca5ab5eb78772ab", "2018-01-10"
        ),
        git.GitTag(
            "project-0.9.2", "1541f54503d2e1cf39bd777c0ca5ab5eb78772ba", "2017-11-11"
        ),
        git.GitTag(
            "ignored-0.9.1", "46e9032e18a819e466618c7a014bcb0e9981af9e", "2017-11-11"
        ),
    ]

    rules = changelog.TagRules(
        scheme=Pep440,
        tag_format="v$version",
        legacy_tag_formats=["$version", "project-${version}"],
        ignored_tag_formats=[
            "known",
            "ignored-${version}",
            "star-*",
            "*-${major}.${minor}",
        ],
    )

    version_tags = rules.get_version_tags(tags, warn=True)
    assert {t.name for t in version_tags} == {
        "v1.1.0",
        "v1.0.0",
        "1.0.0b2",
        "project-0.9.3",
        "project-0.9.2",
    }

    captured = capsys.readouterr()
    assert captured.err.count("Invalid version tag:") == 2
    assert captured.err.count("not-a-version") == 2


@pytest.mark.usefixtures("in_repo_root")
def test_changelog_file_name_from_args_and_config():
    mock_config = Mock(spec=BaseConfig)
    mock_path = Mock(spec=Path)
    mock_path.parent = Path("/my/project")
    mock_config.path = mock_path
    mock_config.settings = {
        "name": "cz_conventional_commits",
        "changelog_file": "CHANGELOG.md",
        "encoding": "utf-8",
        "changelog_start_rev": "v1.0.0",
        "tag_format": "$version",
        "legacy_tag_formats": [],
        "ignored_tag_formats": [],
        "incremental": True,
        "changelog_merge_prerelease": True,
    }

    args = {
        "file_name": "CUSTOM.md",
        "incremental": None,
        "dry_run": False,
        "unreleased_version": "1.0.1",
    }
    changelog = Changelog(mock_config, args)
    assert os.path.normpath(changelog.file_name) == os.path.normpath(
        os.path.join("/my/project", "CUSTOM.md")
    )

    args = {"incremental": None, "dry_run": False, "unreleased_version": "1.0.1"}
    changelog = Changelog(mock_config, args)
    assert os.path.normpath(changelog.file_name) == os.path.normpath(
        os.path.join("/my/project", "CHANGELOG.md")
    )



import io
import sys
from typing import Any

from termcolor import colored

if sys.platform == "win32":
    if isinstance(sys.stdout, io.TextIOWrapper) and sys.version_info >= (3, 7):
        sys.stdout.reconfigure(encoding="utf-8")


def write(value: str, *args: object) -> None:
    """Intended to be used when value is multiline."""
    print(value, *args)


def line(value: str, *args: object, **kwargs: Any) -> None:
    """Wrapper in case I want to do something different later."""
    print(value, *args, **kwargs)


def error(value: str) -> None:
    message = colored(value, "red")
    line(message, file=sys.stderr)


def success(value: str) -> None:
    message = colored(value, "green")
    line(message)


def info(value: str) -> None:
    message = colored(value, "blue")
    line(message)


def diagnostic(value: str) -> None:
    line(value, file=sys.stderr)


def warn(value: str) -> None:
    message = colored(value, "magenta")
    line(message, file=sys.stderr)
